# SPDX-License-Identifier: GPL-3.0-only
from enum import IntEnum

from gi.repository import GLib

QUARK = GLib.quark_from_string("secrets")


class ErrorType(IntEnum):
    UNSUPPORTED_FORMAT = 0
    UNKNOWN = 1
    CREDENTIALS_ERROR = 2

    def message(self):
        match self:
            case ErrorType.UNSUPPORTED_FORMAT:
                return "The kdb Format is not Supported"
            case ErrorType.UNKNOWN:
                return "Unknown error"
            case ErrorType.CREDENTIALS_ERROR:
                return "Invalid credentials"
            case _other:
                msg = "Unrecognized error"
                raise ValueError(msg)


def generic_error(msg):
    return GLib.Error.new_literal(QUARK, msg, ErrorType.UNKNOWN)


def error(type_):
    return GLib.Error.new_literal(QUARK, type_.message(), type_)



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import typing
from gettext import gettext as _

from gi.repository import Adw, GLib, GObject, Gtk

from gsecrets.safe_element import EntryColor, SafeEntry

if typing.TYPE_CHECKING:
    from gsecrets.unlocked_database import UnlockedDatabase


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/entry_row.ui")
class EntryRow(Adw.Bin):
    __gtype_name__ = "EntryRow"

    _prefix_stack = Gtk.Template.Child()
    _entry_icon = Gtk.Template.Child()
    _selection_checkbox = Gtk.Template.Child()
    _entry_copy_otp_button = Gtk.Template.Child()
    _entry_copy_user_button = Gtk.Template.Child()
    _entry_copy_pass_button = Gtk.Template.Child()

    _safe_entry = None

    title = GObject.Property(type=str, default="")
    subtitle = GObject.Property(type=str, default="")

    def __init__(self, database: UnlockedDatabase) -> None:
        super().__init__()

        self._signals = GObject.SignalGroup.new(SafeEntry)
        self._bindings = GObject.BindingGroup.new()

        self._bindings.bind(
            "icon-name",
            self._entry_icon,
            "icon-name",
            GObject.BindingFlags.SYNC_CREATE,
        )
        self._bindings.bind(
            "selected",
            self._selection_checkbox,
            "active",
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE,
        )
        self._bindings.bind(
            "sensitive",
            self,
            "sensitive",
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE,
        )

        self._signals.connect_closure(
            "notify::name",
            self._on_entry_name_changed,
            False,
        )
        self._signals.connect_closure(
            "notify::username",
            self._on_entry_username_changed,
            False,
        )
        self._signals.connect_closure(
            "notify::password",
            self._on_entry_password_changed,
            False,
        )
        self._signals.connect_closure(
            "notify::otp",
            self._on_entry_opt_token_changed,
            False,
        )
        self._signals.connect_closure(
            "notify::color",
            self._on_entry_color_changed,
            False,
        )
        self._signals.connect_closure(
            "notify::expired",
            self._on_entry_notify_expired,
            False,
        )
        self._signals.connect_closure(
            "notify::selected",
            self._on_selected_notify,
            False,
        )

        self.unlocked_database = database
        self.db_manager = database.database_manager

        self.unlocked_database.connect(
            "notify::selection-mode",
            self._on_selection_mode_notify,
        )

    @GObject.Property(type=SafeEntry)
    def safe_entry(self):
        return self._safe_entry

    @safe_entry.setter  # type: ignore
    def safe_entry(self, element):
        self._safe_entry = element

        self._signals.props.target = element
        self._bindings.props.source = element

        self._on_entry_name_changed(element, None)
        self._on_entry_username_changed(element, None)
        self._on_entry_password_changed(element, None)
        self._on_entry_opt_token_changed(element, None)
        self._on_entry_color_changed(element, None)
        self._on_entry_notify_expired(element, None)
        self._on_selection_mode_notify(self.unlocked_database, None)

    def _on_entry_notify_expired(self, safe_entry, _gparam):
        if safe_entry.expired and safe_entry.props.expires:
            self.add_css_class("strikethrough")
        else:
            self.remove_css_class("strikethrough")

    @Gtk.Template.Callback()
    def _on_entry_row_button_pressed(
        self,
        gesture: Gtk.GestureClick,
        _n_press: int,
        _event_x: float,
        _event_y: float,
    ) -> None:
        # pylint: disable=too-many-arguments
        gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        db_view: UnlockedDatabase = self.unlocked_database
        db_view.start_database_lock_timer()

        if db_view.props.selection_mode:
            selected = self._safe_entry.selected  # type: ignore
            self._safe_entry.selected = not selected  # type: ignore
        else:
            db_view.props.selection_mode = True
            self._safe_entry.selected = True  # type: ignore

    def _on_selected_notify(self, entry, _pspec):
        self.unlocked_database.start_database_lock_timer()

        if entry.props.selected:
            self.unlocked_database.add_selection(entry)
        else:
            self.unlocked_database.remove_selection(entry)

    @Gtk.Template.Callback()
    def on_entry_copy_pass_button_clicked(self, _button):
        self.unlocked_database.send_to_clipboard(
            self._safe_entry.props.password,
            _("Password copied"),
        )

    @Gtk.Template.Callback()
    def on_entry_copy_user_button_clicked(self, _button):
        self.unlocked_database.send_to_clipboard(
            self._safe_entry.props.username,
            _("Username copied"),
        )

    @Gtk.Template.Callback()
    def on_entry_copy_otp_button_clicked(self, _button):
        if token := self._safe_entry.otp_token():
            self.unlocked_database.send_to_clipboard(
                token,
                _("One-Time Password copied"),
            )

    def _on_entry_name_changed(
        self,
        safe_entry: SafeEntry,
        _value: GObject.ParamSpec,
    ) -> None:
        entry_name = GLib.markup_escape_text(safe_entry.props.name)
        if entry_name:
            self.remove_css_class("italic-title")
            self.props.title = entry_name
        else:
            self.add_css_class("italic-title")
            self.props.title = _("Title not Specified")

    def _on_entry_username_changed(
        self,
        safe_entry: SafeEntry,
        _value: GObject.ParamSpec,
    ) -> None:
        entry_username = GLib.markup_escape_text(safe_entry.props.username)
        self.props.subtitle = entry_username
        self._entry_copy_user_button.set_visible(bool(entry_username))

    def _on_entry_password_changed(
        self,
        safe_entry: SafeEntry,
        _value: GObject.ParamSpec,
    ) -> None:
        entry_password = GLib.markup_escape_text(safe_entry.props.password)
        self._entry_copy_pass_button.set_visible(bool(entry_password))

    def _on_entry_opt_token_changed(
        self,
        safe_entry: SafeEntry,
        _value: GObject.ParamSpec,
    ) -> None:
        entry_otp_token = safe_entry.otp_token()
        if entry_otp_token:
            self._entry_copy_otp_button.props.visible = True
        else:
            self._entry_copy_otp_button.props.visible = False

    def _on_entry_color_changed(
        self,
        safe_entry: SafeEntry,
        _value: GObject.ParamSpec,
    ) -> None:
        # Clear current style
        for color in EntryColor:
            css_class = color.css_class()
            self._entry_icon.remove_css_class(css_class)

        color = safe_entry.props.color
        css_class = EntryColor(color).css_class()
        self._entry_icon.add_css_class(css_class)

    def _on_selection_mode_notify(self, unlocked_db, _pspec):
        selection_mode = unlocked_db.props.selection_mode

        if selection_mode:
            visible_child = self._selection_checkbox
        else:
            visible_child = self._entry_icon

        self._prefix_stack.props.visible_child = visible_child

    @Gtk.Template.Callback()
    def _on_long_press_gesture_pressed(self, gesture, _x, _y):
        gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        self.unlocked_database.props.selection_mode = True
        self._safe_entry.props.selected = not self._safe_entry.props.selected



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import logging
from gettext import gettext as _
from pathlib import Path

from gi.repository import Adw, Gio, GLib, Gtk


class SavingConflictDialog(Adw.AlertDialog):
    __gtype_name__ = "SavingConflictDialog"

    def __init__(self, window, db_manager, save_callback):
        super().__init__()

        self.save_callback = save_callback
        self.db_manager = db_manager
        self.window = window

        # TRANSLATORS Warning dialog to resolve file saving conflicts.
        self.props.heading = _("Conflicts While Saving")
        self.props.body = _(
            # TRANSLATORS Warning Dialog to resolve saving conflicts. \n is a new line.
            "The safe was modified from somewhere else. Saving will overwrite their version of the safe with our current version.\n\n You can also make a backup of their version of the safe.",  # noqa: E501
        )

        gfile = Gio.File.new_for_path(db_manager.path)
        file_name = Path(gfile.get_basename()).stem

        self.add_response("cancel", _("_Cancel"))
        # TRANSLATORS backup and save current safe.
        self.add_response("backup", _("_Back up and Save"))
        self.add_response("save", _("_Save"))
        self.set_response_appearance("save", Adw.ResponseAppearance.DESTRUCTIVE)

        self.connect("response::save", self._on_response_save)
        self.connect("response::backup", self._on_response_backup, file_name)

    def _on_response_save(self, _dialog, _response):
        self.db_manager.save_async(self.save_callback)

    def _on_response_backup(self, _message_dialog, _response, file_name):
        dialog = Gtk.FileDialog.new()
        dialog.props.title = _("Save Backup")
        dialog.props.initial_name = f"{file_name}-backup.kdbx"

        dialog.save(self.window, None, self._on_filechooser_response)

    def _on_copy_backup(self, gfile, result):
        try:
            gfile.copy_finish(result)
        except GLib.Error:
            logging.exception("Could not backup safe")
            self.window.send_notification(_("Could not backup safe"))
        else:
            self.db_manager.save_async(self.save_callback)

    def _on_filechooser_response(self, dialog, result):
        try:
            dest = dialog.save_finish(result)
        except GLib.Error as err:
            if not err.matches(Gtk.DialogError.quark(), Gtk.DialogError.DISMISSED):
                logging.exception("Could not save file")
        else:
            gfile = Gio.File.new_for_path(self.db_manager.path)
            if GLib.check_version(2, 81, 0) is None:
                gfile.copy_async(
                    dest,
                    Gio.FileCopyFlags.OVERWRITE,
                    GLib.PRIORITY_DEFAULT,
                    None,
                    None,
                    self._on_copy_backup,
                )
            else:
                gfile.copy_async(
                    dest,
                    Gio.FileCopyFlags.OVERWRITE,
                    GLib.PRIORITY_DEFAULT,
                    None,
                    None,
                    None,
                    self._on_copy_backup,
                )



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import logging
import typing
from gettext import gettext as _
from pathlib import Path

from gi.repository import Adw, Gio, GLib, Gtk

from gsecrets import const

if typing.TYPE_CHECKING:
    from pykeepass.attachment import Attachment

    from gsecrets.entry_page import EntryPage  # pylint: disable=ungrouped-imports


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/attachment_warning_dialog.ui")
class AttachmentWarningDialog(Adw.AlertDialog):
    __gtype_name__ = "AttachmentWarningDialog"

    def __init__(self, entry_page: EntryPage, attachment: Attachment) -> None:
        """Dialog to confirm opening an attachment.

        :param entry_page: entry page
        """
        super().__init__()

        self.__unlocked_database = entry_page.unlocked_database
        self.__attachment = attachment

    @Gtk.Template.Callback()
    def _on_proceed(self, _dialog, _response):
        attachment = self.__attachment
        u_db = self.__unlocked_database
        self.__open_tmp_file(
            u_db.database_manager.db.binaries[attachment.id],
            attachment.filename,
        )

    def __open_tmp_file(self, bytes_buffer, filename):
        cache_dir = Path(GLib.get_user_cache_dir()) / const.SHORT_NAME / "tmp"
        file_path = cache_dir / filename
        window = self.__unlocked_database.window
        try:
            cache_dir.mkdir(parents=True, exist_ok=True)
        except FileExistsError:
            logging.exception("Could not create tmp dir")
            return

        gfile = Gio.File.new_for_path(str(file_path))

        contents = GLib.Bytes.new(bytes_buffer)
        gfile.replace_contents_bytes_async(
            contents,
            None,
            False,
            Gio.FileCreateFlags.PRIVATE | Gio.FileCreateFlags.REPLACE_DESTINATION,
            None,
            _callback,
            window,
        )


def _callback(gfile, result, window):
    try:
        gfile.replace_contents_finish(result)
    except GLib.Error:
        logging.exception("Could not load attachment")
        window.send_notification(_("Could not load attachment"))
    else:
        launcher = Gtk.FileLauncher.new(gfile)
        launcher.launch(window, None, _on_launch, window)


def _on_launch(launcher, result, window):
    try:
        launcher.launch_finish(result)
    except GLib.Error:
        logging.exception("Could not launch attachment file")
        window.send_notification(_("Could not load attachment"))



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from gi.repository import Adw, Gtk


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/unlocked_headerbar.ui")
class UnlockedHeaderBar(Adw.Bin):
    __gtype_name__ = "UnlockedHeaderBar"

    _go_back_button = Gtk.Template.Child()
    _selection_button = Gtk.Template.Child()

    def __init__(self, unlocked_database):
        """HeaderBar of an UnlockedDatabase.

        :param UnlockedDatabase unlocked_database: unlocked_database
        """
        super().__init__()

        self._unlocked_database = unlocked_database
        unlocked_database.connect(
            "notify::current-element",
            self._on_current_element_notify,
        )

        self._go_back_button.connect("clicked", self.on_go_back_button_clicked)
        self._selection_button.connect("clicked", self._on_selection_button_clicked)

    def _on_selection_button_clicked(self, _button: Gtk.Button) -> None:
        self._unlocked_database.props.selection_mode = True

    def on_go_back_button_clicked(self, _button):
        group = self._unlocked_database.current_element
        self._unlocked_database.show_browser_page(group.parentgroup)

    def _on_current_element_notify(self, unlocked_db, _pspec):
        visible = not unlocked_db.current_element.is_root_group
        self._go_back_button.props.visible = visible



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import logging
import typing
from gettext import gettext as _

from gi.repository import Adw, Gio, GLib, Gtk

from gsecrets.attachment_delete_dialog import AttachmentDeleteDialog

if typing.TYPE_CHECKING:
    from pykeepass.attachment import Attachment

    from gsecrets.add_list_box import AddListBox
    from gsecrets.safe_element import SafeEntry
    from gsecrets.unlocked_database import UnlockedDatabase


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/attachment_entry_row.ui")
class AttachmentEntryRow(Adw.ActionRow):
    __gtype_name__ = "AttachmentEntryRow"

    download_button = Gtk.Template.Child()
    delete_button = Gtk.Template.Child()

    def __init__(
        self,
        unlocked_database: UnlockedDatabase,
        entry: SafeEntry,
        attachment: Attachment,
        listbox: AddListBox,
    ) -> None:
        super().__init__()

        self.__unlocked_database = unlocked_database

        self.entry = entry
        self.attachment = attachment
        self._listbox = listbox

        self.set_title(attachment.filename)
        # TODO Display mime type in subtitle

    @Gtk.Template.Callback()
    def _on_download_button_clicked(self, _button):
        window = self.get_root()
        dialog = Gtk.FileDialog.new()
        # NOTE: Filechooser title for downloading an attachment
        dialog.props.title = _("Save Attachment")
        dialog.props.initial_name = self.attachment.filename

        dialog.save(window, None, self._on_save_filechooser_response)

    @Gtk.Template.Callback()
    def _on_delete_button_clicked(self, _button):
        def on_delete():
            self.entry.delete_attachment(self.attachment)
            self._listbox.remove(self)

        AttachmentDeleteDialog(
            self.__unlocked_database, self.attachment, on_delete
        ).present(self.get_root())

    def _replace_contents_callback(self, gfile, result):
        try:
            gfile.replace_contents_finish(result)
        except GLib.Error:
            logging.exception("Could not store attachment")
            window = self.get_root()
            window.send_notification(_("Could not store attachment"))

    def _on_save_filechooser_response(self, dialog, result):
        try:
            gfile = dialog.save_finish(result)
        except GLib.Error as err:
            if not err.matches(Gtk.DialogError.quark(), Gtk.DialogError.DISMISSED):
                logging.exception("Could not save file")
        else:
            bytes_buffer = self.entry.get_attachment_content(self.attachment)
            gbytes = GLib.Bytes.new(bytes_buffer)

            gfile.replace_contents_bytes_async(
                gbytes,
                None,
                False,
                Gio.FileCreateFlags.PRIVATE | Gio.FileCreateFlags.REPLACE_DESTINATION,
                None,
                self._replace_contents_callback,
            )



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from gettext import gettext as _

from gi.repository import Adw, GObject, Gtk

from gsecrets.utils import format_time


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/history_row.ui")
class HistoryRow(Adw.ActionRow):
    __gtype_name__ = "HistoryRow"

    visibility_button = Gtk.Template.Child()

    _reveal: bool = False

    def __init__(self, history_entry, window):
        super().__init__()

        self.history_entry = history_entry
        self.list_model = window.list_model
        self.password = history_entry.password
        self.safe_entry = window.safe_entry
        self.unlocked_database = window.unlocked_database
        self.window = window

        self.props.subtitle = format_time(history_entry.mtime)
        self.props.title = "•" * len(self.password)

        self.install_property_action("historyrow.reveal", "reveal")

    @GObject.Property(type=bool, default=False)
    def reveal(self) -> bool:
        return self._reveal

    @reveal.setter  # type: ignore
    def reveal(self, reveal) -> bool:
        button = self.visibility_button
        if reveal:
            button.props.icon_name = "eye-not-looking-symbolic"
            button.props.tooltip_text = _("Hide Text")
            self.props.title = self.password
        else:
            button.props.icon_name = "eye-open-negative-filled-symbolic"
            button.props.tooltip_text = _("Show Text")
            self.props.title = "•" * len(self.password)

        self._reveal = reveal

    @Gtk.Template.Callback()
    def _on_copy_button_clicked(self, _button):
        self.unlocked_database.send_to_clipboard(self.password)

    @Gtk.Template.Callback()
    def _on_delete_button_clicked(self, _button):
        self.safe_entry.delete_history(self.history_entry)
        found, pos = self.list_model.find(self.history_entry)
        if found:
            self.list_model.remove(pos)

        if self.list_model.get_n_items() == 0:
            self.window.close()



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import io
import json
import logging
from pathlib import Path
from time import perf_counter

from gi.repository import Gio, GLib, GObject
from pykeepass import PyKeePass
from pykeepass.exceptions import CredentialsError

import gsecrets.config_manager as config
from gsecrets.err import ErrorType, error, generic_error
from gsecrets.recent_manager import RecentManager
from gsecrets.safe_element import SafeEntry, SafeGroup
from gsecrets.utils import LazyValue, compare_passwords

# GTK uses G_PRIORITY_HIGH_IDLE + 10 for resizing operations, and
# G_PRIORITY_HIGH_IDLE + 20 for redrawing operations. We use a slightly lower
# priority.
PRIORITY = GLib.PRIORITY_HIGH_IDLE + 25
BATCH_SIZE: int = 10


class DatabaseManager(GObject.Object):
    # pylint: disable=too-many-public-methods
    # pylint: disable=too-many-instance-attributes
    """Implements database functionality that is independent of the UI.

    Useful attributes:
     .path: str containing the filepath of the database
     .is_dirty: bool telling whether the database is in a dirty state

    Group objects are of type `pykeepass.group.Group`
    Entry objects are of type `pykeepass.entry.Entry`
    Instances of both have useful attributes:
    .uuid: a `uuid.UUID` object
    """

    # self.db contains a `PyKeePass` database
    _elements_loaded = False
    password_try = ""
    _is_dirty = False  # Does the database need saving?
    save_running = False
    _opened = False

    # Used for checking if there are changes in the file.
    file_size: int | None = None
    file_mtime: int | None = None

    # Only used for setting the credentials to their actual values in case of
    # errors.
    old_password: str = ""
    composition_key: bytes | None = None

    trash_bin: SafeGroup | None = None
    root: SafeGroup | None = None

    locked = GObject.Property(type=bool, default=False)
    is_dirty = GObject.Property(type=bool, default=False)

    # To be emitted when the elements list model needs to be re sorted.
    sorting_changed = GObject.Signal(arg_types=(bool,))

    def __init__(self, key_providers: list, database_path: str) -> None:
        """Initialize the database handling logic.

        :param str database_path: The database path
        """
        super().__init__()

        self.entries = Gio.ListStore.new(SafeEntry)
        self.groups = Gio.ListStore.new(SafeGroup)

        self._key_providers = key_providers
        self._path = database_path
        self.db: PyKeePass = None

    def unlock_async(
        self,
        password: str,
        composition_key: bytes | None = None,
        callback: Gio.AsyncReadyCallback = None,
    ) -> None:
        """Unlocks and opens a safe.

        This pykeepass to open a safe database. If the database cannot
        be opened, an exception is raised.

        :param str password: password to use or an empty string
        :param str composition_key: composition_key as bytes
        :param GAsyncReadyCallback: callback run after the unlock operation ends
        """
        self._opened = False

        def unlock_task(task, _source_object, _task_data, _cancellable):
            if Path(self._path).suffix == ".kdb":
                # NOTE kdb is a an older format for Keepass databases.
                err = error(ErrorType.UNSUPPORTED_FORMAT)
                task.return_error(err)
                return

            try:
                db = PyKeePass(
                    self.path,
                    password,
                    io.BytesIO(composition_key) if composition_key else None,
                )
            except CredentialsError:
                err = error(ErrorType.CREDENTIALS_ERROR)
                task.return_error(err)
            except Exception as err:  # pylint: disable=broad-except
                err = generic_error(str(err))
                task.return_error(err)
            else:
                self.composition_key = composition_key
                self._update_file_monitor()
                task.return_value(db)

        task = Gio.Task.new(self, None, callback)
        task.run_in_thread(unlock_task)

    def unlock_finish(self, result):
        """Finish unlocking.

        Can raise errors.
        """
        _success, db = result.propagate_value()

        self.db = db
        self._opened = True
        logging.debug("Opening of safe %s was successful", self.path)

        if not self._elements_loaded:
            now = perf_counter()
            GLib.idle_add(self._load_groups, now, priority=PRIORITY)
            GLib.idle_add(self._load_entries, now, priority=PRIORITY)

            self._elements_loaded = True

        self.notify("description")
        self.notify("name")

    def _load_groups(self, then):
        total = len(self.db.groups)
        n_loaded = self.groups.get_n_items()

        if n_loaded == total:
            elapsed = perf_counter() - then
            logging.debug("Finished loading %s groups in %.2fs", n_loaded, elapsed)
            return GLib.SOURCE_REMOVE

        to_load = min(BATCH_SIZE, total - n_loaded)

        groups = [
            SafeGroup(self, g) for g in self.db.groups[n_loaded : n_loaded + to_load]
        ]
        self.groups.splice(n_loaded, 0, groups)

        return GLib.SOURCE_CONTINUE

    def _load_entries(self, then):
        total = len(self.db.entries)
        n_loaded = self.entries.get_n_items()

        if n_loaded == total:
            elapsed = perf_counter() - then
            logging.debug("Finished loading %s entries in %.2fs", n_loaded, elapsed)
            return GLib.SOURCE_REMOVE

        to_load = min(BATCH_SIZE, total - n_loaded)

        entries = [
            SafeEntry(self, e) for e in self.db.entries[n_loaded : n_loaded + to_load]
        ]
        self.entries.splice(n_loaded, 0, entries)

        return GLib.SOURCE_CONTINUE

    #
    # Database Modifications
    #

    def save_async(self, callback: Gio.AsyncReadyCallback) -> None:
        """Write all changes to database.

        This consists in two parts, we first save it to a byte stream in memory
        and then we write the contents of the stream into the database using
        Gio. This is done in this way since pykeepass save is not atomic,
        whereas GFile operations are.

        Note that certain operations in pykeepass can fail at the middle of the
        operation, nuking the database in the process.
        """
        task = Gio.Task.new(self, None, callback)
        task.run_in_thread(self._save_task)

    def _save_task(self, task, _obj, _data, _cancellable):
        if self.save_running:
            task.return_boolean(False)
            logging.debug("Save already running")
            return

        if not self.is_dirty:
            task.return_boolean(False)
            logging.debug("Safe is not dirty")
            return

        logging.debug("Saving database %s", self.path)
        self.save_running = True

        with io.BytesIO() as buf:
            try:
                self.db.save(buf)
            except Exception as err:  # pylint: disable=broad-except
                err = generic_error(str(err))
                task.return_error(err)
            else:
                flags = Gio.FileCreateFlags.NONE
                gfile = Gio.File.new_for_path(self._path)
                contents = buf.getvalue()
                try:
                    gfile.replace_contents(contents, None, False, flags, None)
                except GLib.Error as err:  # pylint: disable=broad-except
                    task.return_error(err)
                else:
                    self._update_file_monitor()
                    task.return_boolean(True)

    def save_finish(self, result: Gio.AsyncResult) -> bool:
        """Finish save_async.

        Returns whether the safe was saved. Can raise GLib.Error.
        """
        self.save_running = False
        is_saved = result.propagate_boolean()

        self.is_dirty = False
        if is_saved:
            logging.debug("Database %s saved successfully", self.path)

        return is_saved

    def set_credentials_async(
        self,
        password: str,
        composition_key: bytes | None = None,
        callback: Gio.AsyncReadyCallback = None,
    ) -> None:
        """Set credentials for safe.

        It does almost the same as save_async, with the difference that
        correctly handles errors, it won't leave the database in a state where
        its fields have the incorrect values.
        """

        def set_credentials_task(task, obj, data, cancellable):
            self.old_password = self.password
            self.password = password

            self.composition_key = composition_key
            if composition_key:
                self.db.keyfile = io.BytesIO(composition_key)
            else:
                self.db.keyfile = None

            self.is_dirty = True

            self._save_task(task, obj, data, cancellable)

        task = Gio.Task.new(self, None, callback)
        task.run_in_thread(set_credentials_task)

    def set_credentials_finish(self, result):
        self.save_running = False
        try:
            is_saved = result.propagate_boolean()
        except GLib.Error:
            self.password = self.old_password

            raise

        self.is_dirty = False
        if is_saved:
            logging.debug("Credentials changed successfully")

        return is_saved

    def add_to_history(self) -> None:
        gfile = Gio.File.new_for_path(self._path)
        uri = gfile.get_uri()

        recents = RecentManager()
        recents.add_item(gfile)

        # Set last opened database.
        config.set_last_opened_database(uri)

        if not config.get_remember_composite_key():
            return

        keyprovider_pairs = config.get_last_used_key_provider()

        data = {}
        for key_provider in self._key_providers:
            if key_provider.available and key_provider.key:
                provider_config = key_provider.config()
                data[type(key_provider).__name__] = provider_config

        keyprovider_pairs[uri] = json.dumps(data)
        config.set_last_used_key_provider(keyprovider_pairs)

    def _update_file_monitor(self):
        """Update the modified time and size of the database.

        This is a blocking operation.
        """
        gfile = Gio.File.new_for_path(self._path)
        attributes = (
            f"{Gio.FILE_ATTRIBUTE_STANDARD_SIZE},{Gio.FILE_ATTRIBUTE_TIME_MODIFIED}"
        )
        try:
            info = gfile.query_info(attributes, Gio.FileQueryInfoFlags.NONE, None)
        except GLib.Error:
            logging.exception("Could not read file size")
        else:
            self.file_size = info.get_size()
            self.file_mtime = info.get_modification_date_time().to_unix()

    def check_file_changes_async(self, callback: Gio.AsyncReadyCallback) -> None:
        task = Gio.Task.new(self, None, callback)
        task.run_in_thread(self._check_file_changes_task)

    def _check_file_changes_task(self, task, _obj, _data, _cancellable):
        gfile = Gio.File.new_for_path(self._path)
        attributes = (
            f"{Gio.FILE_ATTRIBUTE_STANDARD_SIZE},{Gio.FILE_ATTRIBUTE_TIME_MODIFIED}"
        )
        try:
            info = gfile.query_info(attributes, Gio.FileQueryInfoFlags.NONE, None)
        except GLib.Error as err:
            err = generic_error(str(err))
            task.return_error(err)
        else:
            if (
                self.file_size != info.get_size()
                or self.file_mtime != info.get_modification_date_time().to_unix()
            ):
                task.return_boolean(True)
            else:
                task.return_boolean(False)

    def check_file_changes_finish(self, result: Gio.AsyncResult) -> bool:
        """Finish check_file_changes_async.

        Returns whether the file was changed, can raise GLib.Error.
        """
        return result.propagate_boolean()

    @property
    def password(self) -> str:
        """Get the current password or '' if not set."""
        if db := self.db:
            return db.password or ""

        return ""

    @password.setter
    def password(self, new_password: str | None) -> None:
        """Set database password (None if a old_composition key is used)."""
        self.db.password = new_password
        self.is_dirty = True

    @GObject.Property(type=str, default="")
    def name(self) -> str:
        """Get the database name or '' if not set."""
        return self.db.database_name or ""

    @name.setter  # type: ignore
    def name(self, new_name: str | None) -> None:
        """Set database name."""
        self.db.database_name = new_name
        self.is_dirty = True

    @GObject.Property(type=str, default="")
    def description(self) -> str:
        """Get the database description or '' if not set."""
        return self.db.database_description or ""

    @description.setter  # type: ignore
    def description(self, new_description: str | None) -> None:
        """Set database description."""
        self.db.database_description = new_description
        self.is_dirty = True

    @property
    def default_username(self) -> str:
        """Get the default username or '' if not set."""
        return self.db.default_username or ""

    @default_username.setter
    def default_username(self, new_username: str | None) -> None:
        """Set default username."""
        self.db.default_username = new_username
        self.is_dirty = True

    #
    # Read Database
    #

    # Check if entry with title in group exists
    def check_entry_in_group_exists(self, title, group):
        entry = self.db.find_entries(
            title=title,
            group=group,
            recursive=False,
            history=False,
            first=True,
        )
        return entry is not None

    #
    # Database creation methods
    #

    # Set the first password entered by the user (for comparing reasons)
    def set_password_try(self, password):
        self.password_try = password

    def compare_passwords(self, password2: str) -> bool:
        """Compare the first password entered by the user with the second one.

        It also does not allow empty passwords.
        :returns: True if passwords match and are non-empty.
        """
        return compare_passwords(self.password_try, password2)

    def parent_checker(self, current_group, moved_group):
        """Return True if moved_group is an ancestor of current_group."""
        # recursively invoke ourself until we reach the root group
        if current_group.is_root_group:
            return False
        if current_group.uuid == moved_group.uuid:
            return True
        return self.parent_checker(current_group.parentgroup, moved_group)

    @property
    def version(self):
        """Returns the database version."""
        return self.db.version

    @property
    def opened(self) -> bool:
        return self._opened

    @property
    def path(self) -> str:
        return self._path

    def get_salt_as_lazy(self) -> LazyValue[bytes]:
        path = self.path
        return LazyValue(lambda: DatabaseManager.salt_from_path(path))

    # This is static so as there is no need to carry around a
    # DatabaseManager object in another thread
    @staticmethod
    def salt_from_path(path: str) -> bytes:
        db = PyKeePass(path, None, None, decrypt=False)
        return db.database_salt



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import logging
import typing
from gettext import gettext as _

import validators
from gi.repository import Adw, GLib, GObject, Gtk

from gsecrets.attachment_warning_dialog import AttachmentWarningDialog
from gsecrets.color_widget import ColorEntryRow
from gsecrets.safe_element import ICONS, SafeEntry
from gsecrets.widgets.add_attribute_dialog import AddAttributeDialog
from gsecrets.widgets.attachment_entry_row import AttachmentEntryRow
from gsecrets.widgets.attribute_entry_row import AttributeEntryRow
from gsecrets.widgets.entry_page_icon import EntryPageIcon
from gsecrets.widgets.entry_page_tag import EntryPageTag
from gsecrets.widgets.history_window import HistoryWindow
from gsecrets.widgets.notes_dialog import NotesDialog
from gsecrets.widgets.protected_attribute_entry_row import ProtectedAttributeEntryRow

if typing.TYPE_CHECKING:
    from gi.repository import Gio
    from pykeepass.attachment import Attachment

    from gsecrets.safe_element import SafeElement
    from gsecrets.unlocked_database import UnlockedDatabase


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/entry_page.ui")
class EntryPage(Adw.Bin):
    # pylint: disable=too-many-public-methods

    __gtype_name__ = "EntryPage"

    title_entry_row = Gtk.Template.Child()
    url_entry_row = Gtk.Template.Child()

    credentials_group = Gtk.Template.Child()

    otp_error_revealer = Gtk.Template.Child()
    otp_preferences_group = Gtk.Template.Child()
    otp_progress_icon = Gtk.Template.Child()
    otp_secret_entry_row = Gtk.Template.Child()
    otp_token_row = Gtk.Template.Child()

    notes_preferences_group = Gtk.Template.Child()
    notes_property_value_entry = Gtk.Template.Child()

    color_property_bin = Gtk.Template.Child()

    icon_entry_box = Gtk.Template.Child()

    _tags_group = Gtk.Template.Child()
    _tag_entry_box = Gtk.Template.Child()

    attachments_preferences_group = Gtk.Template.Child()
    attachment_list_box = Gtk.Template.Child()

    attribute_list_box = Gtk.Template.Child()
    attributes_preferences_group = Gtk.Template.Child()

    expiration_date_preferences_group = Gtk.Template.Child()
    expiration_date_row = Gtk.Template.Child()

    _window_title = Gtk.Template.Child()

    show_all_preferences_group = Gtk.Template.Child()

    otp_timer_handler: int | None = None

    safe_entry = GObject.Property(type=SafeEntry)

    _tags: list[EntryPageTag] = []

    def __init__(
        self,
        u_d: UnlockedDatabase,
        entry: SafeElement,
        add_all: bool,
    ) -> None:
        # Setup actions, must be done before initializing the template.
        self.install_action("entry.copy_password", None, self._on_copy_action)
        self.install_action("entry.copy_user", None, self._on_copy_action)
        self.install_action("entry.copy_otp", None, self._on_copy_action)
        self.install_action("entry.copy_url", None, self._on_copy_action)
        self.install_action("entry.add_attribute", None, self._on_add_attribute)
        self.install_action("entry.add_attachment", None, self._on_add_attachment)
        self.install_action(
            "entry.password_history",
            None,
            self._on_password_history_action,
        )
        self.install_action(
            "entry.save_in_history",
            None,
            self._on_save_in_history_action,
        )

        super().__init__(safe_entry=entry)

        self.unlocked_database = u_d
        self.toggeable_widget_list = [
            self.otp_preferences_group,
            self.notes_preferences_group,
            self.attachments_preferences_group,
            self.attributes_preferences_group,
            self.expiration_date_preferences_group,
            self._tags_group,
        ]

        self.insert_entry_properties_into_listbox(add_all)

        safe_entry: SafeEntry = self.props.safe_entry
        safe_entry.updated.connect(self._on_safe_entry_updated)

        safe_entry.history_saved.connect(self._on_history_saved)
        if not safe_entry.history:
            self.action_set_enabled("entry.password_history", False)

        safe_entry.connect("notify::name", self._on_name_notify)
        self._set_title(safe_entry)

    def _set_title(self, entry):
        if entry.parentgroup.is_root_group:
            title = entry.name
        else:
            title = f"{entry.parentgroup.name} ⟩ {entry.name}"

        self._window_title.props.title = title

    def _on_name_notify(self, entry, _pspec):
        self._set_title(entry)

    def do_unroot(self) -> None:  # pylint: disable=arguments-differ
        if self.otp_timer_handler is not None:
            GLib.source_remove(self.otp_timer_handler)
            self.otp_timer_handler = None

        Gtk.Widget.do_unroot(self)

    def insert_entry_properties_into_listbox(self, add_all: bool) -> None:
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements
        safe_entry = self.props.safe_entry

        # Create the name_property_row
        safe_entry.bind_property(
            "name",
            self.title_entry_row,
            "text",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )

        # Credentials Group
        self.credentials_group.unlocked_database = self.unlocked_database
        self.credentials_group.props.entry = self.props.safe_entry

        # OTP token
        safe_entry.connect("notify::otp", self.otp_update)
        if safe_entry.otp_token():
            self.otp_update(safe_entry, None)
            self.show_row(self.otp_preferences_group, True, add_all)

        # Url
        safe_entry.bind_property(
            "url",
            self.url_entry_row,
            "text",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )

        # OTP secret
        safe_entry.bind_property(
            "otp",
            self.otp_secret_entry_row,
            "text",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )

        # Notes
        textbuffer = self.notes_property_value_entry.get_buffer()
        safe_entry.bind_property(
            "notes",
            textbuffer,
            "text",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )
        self.notes_property_value_entry.set_buffer(textbuffer)
        self.show_row(self.notes_preferences_group, safe_entry.notes, add_all)

        # Color
        self.color_property_bin.set_child(
            ColorEntryRow(self.unlocked_database, safe_entry),
        )

        # Icons
        entry_icon = safe_entry.props.icon
        for icon_nr, icon in ICONS.items():
            if not icon.visible:
                continue

            btn = EntryPageIcon(icon.name, icon_nr)
            self.icon_entry_box.insert(btn, -1)

            if entry_icon == icon:
                self.icon_entry_box.select_child(btn)

        self.icon_entry_box.connect(
            "selected-children-changed",
            self.on_entry_icon_button_toggled,
        )

        # Tags
        self._update_tags()
        self.show_row(self._tags_group, bool(safe_entry.tags), add_all)

        # Attachments
        for attachment in safe_entry.attachments:
            self.add_attachment_row(attachment)

        self.show_row(
            self.attachments_preferences_group,
            safe_entry.attachments,
            add_all,
        )

        # Attributes
        self.show_row(self.attributes_preferences_group, safe_entry.attributes, add_all)

        def factory(item):
            key, value = item.key, item.value
            if safe_entry.is_attribute_protected(key):
                return ProtectedAttributeEntryRow(safe_entry, key, value).row

            return AttributeEntryRow(safe_entry, key, value)

        self.attribute_list_box.set_model(safe_entry.attributes, factory)

        # Expiration Date
        self.expiration_date_row.props.safe_entry = safe_entry
        self.show_row(
            self.expiration_date_preferences_group,
            safe_entry.expires,
            add_all,
        )

        # Show more row
        for widget in self.toggeable_widget_list:
            if not widget.get_visible():
                self.show_all_preferences_group.props.visible = True
                break

    def _update_tags(self):
        safe_entry = self.props.safe_entry

        for tag in self._tags:
            self._tag_entry_box.remove(tag)

        self._tags = []

        if safe_entry.tags:
            for tag in safe_entry.tags:
                widget = EntryPageTag(tag)
                widget.remove_button.connect(
                    "clicked",
                    self._on_remove_button_clicked,
                    tag,
                )
                self._tag_entry_box.append(widget)
                self._tags.append(widget)

    def _on_remove_button_clicked(self, _button, tag_name):
        safe_entry = self.props.safe_entry
        # Remove
        tags = safe_entry.tags
        tags.remove(tag_name)
        safe_entry.tags = tags or ""

        # Update tag flowbox
        self._update_tags()

    @Gtk.Template.Callback()
    def _on_tag_add_clicked(self, _widget):
        entry = Adw.EntryRow()
        entry.props.activates_default = True
        entry.props.title = _("Enter tag name")

        group = Adw.PreferencesGroup()
        group.add(entry)

        dialog = Adw.AlertDialog.new()
        dialog.set_heading(_("Add New Tag"))
        dialog.add_response("cancel", _("_Cancel"))
        dialog.add_response("add", _("_Add"))
        dialog.set_default_response("add")
        dialog.set_close_response("cancel")
        dialog.set_extra_child(group)

        dialog.connect("response::add", self._on_tag_add_response, entry)
        dialog.present(self.unlocked_database.window)

        entry.grab_focus()

    def _on_tag_add_response(
        self,
        _dialog: Adw.MessageDialog,
        _response: str,
        entry: Adw.EntryRow,
    ) -> None:
        tag_name = entry.props.text
        safe_entry = self.props.safe_entry
        tags = safe_entry.tags or []

        if tag_name in tags:
            window = self.unlocked_database.window
            window.send_notification(_("The tag already exists"))
            return

        tags.append(tag_name)
        safe_entry.tags = tags
        self._update_tags()

    #
    # Events
    #

    def _on_password_history_action(self, widget, _action, _param):
        safe_entry = widget.props.safe_entry

        window = HistoryWindow(safe_entry, widget.unlocked_database)
        window.present(widget)

    def _on_save_in_history_action(self, widget, _action, _param):
        safe_entry = widget.props.safe_entry
        safe_entry.save_history()

        widget.unlocked_database.window.send_notification(_("Saved in history"))

    def _on_history_saved(self, entry):
        if not entry.history:
            self.action_set_enabled("entry.password_history", False)
        else:
            self.action_set_enabled("entry.password_history", True)

    def _on_copy_action(
        self,
        widget: Gtk.Widget,
        action_name: str,
        _parameter: GLib.Variant,
    ) -> None:
        if widget.unlocked_database.props.database_locked:
            return

        if action_name == "entry.copy_user":
            username = widget.credentials_group.username
            widget.unlocked_database.send_to_clipboard(username, _("Username copied"))
        elif action_name == "entry.copy_password":
            widget.credentials_group.copy_password()
        elif action_name == "entry.copy_url":
            url = widget.url_entry_row.props.text
            widget.unlocked_database.send_to_clipboard(url, _("Address copied"))
        elif action_name == "entry.copy_otp":
            safe_entry: SafeEntry = widget.props.safe_entry
            otp_token = safe_entry.otp_token() or ""
            if otp_token == "":
                return

            widget.unlocked_database.send_to_clipboard(
                otp_token,
                _("One-time password copied"),
            )

    @Gtk.Template.Callback()
    def on_show_all_properties_button_clicked(self, _widget):
        self.unlocked_database.start_database_lock_timer()

        for widget in self.toggeable_widget_list:
            widget.set_visible(True)

        self.show_all_preferences_group.props.visible = False

    @Gtk.Template.Callback()
    def on_notes_detach_button_clicked(self, _button):
        self.unlocked_database.start_database_lock_timer()
        safe_entry = self.props.safe_entry
        NotesDialog(self.unlocked_database, safe_entry).present(self)

    def on_entry_icon_button_toggled(self, flowbox):
        if not flowbox.get_selected_children():
            return

        selected_row = flowbox.get_selected_children()[0]
        icon = selected_row.get_name()

        safe_entry = self.props.safe_entry
        safe_entry.props.icon = icon

    def _on_launch(self, launcher, result, window):
        try:
            launcher.launch_finish(result)
        except GLib.Error:
            logging.exception("Could not launch uri")
            window.send_notification(_("Could not open URL"))

    @Gtk.Template.Callback()
    def on_visit_url_button_clicked(self, _button):
        self.unlocked_database.start_database_lock_timer()
        url = self.url_entry_row.props.text
        window = self.unlocked_database.window
        if not validators.url(url):
            url = "https://" + url
            if not validators.url(url):
                window.send_notification(_("The address is not valid"))
                return

        launcher = Gtk.UriLauncher.new(url)
        launcher.launch(window, None, self._on_launch, window)

    @Gtk.Template.Callback()
    def on_otp_copy_button_clicked(self, _button: Gtk.Button) -> None:
        safe_entry: SafeEntry = self.props.safe_entry
        otp_token = safe_entry.otp_token() or ""
        self.unlocked_database.send_to_clipboard(
            otp_token,
            _("One-time password copied"),
        )

    def _on_add_attribute(self, widget, _action_name, _pspec):
        window = widget.unlocked_database.window
        entry = widget.props.safe_entry
        db_manager = widget.unlocked_database.database_manager
        dialog = AddAttributeDialog(db_manager, entry)
        dialog.present(window)

    #
    # Attachment Handling
    #

    @Gtk.Template.Callback()
    def on_attachment_list_box_activated(self, _widget, list_box_row):
        self.unlocked_database.start_database_lock_timer()
        attachment = list_box_row.attachment
        AttachmentWarningDialog(self, attachment).present(self)

    def _on_add_attachment(self, widget, _action_name, _pspec):
        widget.unlocked_database.start_database_lock_timer()

        dialog = Gtk.FileDialog.new()
        # NOTE: Filechooser title for selecting attachment file
        dialog.props.title = _("Select Attachments")
        dialog.props.accept_label = _("_Add")

        dialog.open_multiple(
            widget.unlocked_database.window,
            None,
            widget.on_select_filechooser_response,
        )

    def on_select_filechooser_response(
        self,
        dialog: Gtk.FileDialog,
        result: Gio.AsyncResult,
    ) -> None:
        try:
            files = dialog.open_multiple_finish(result)
        except GLib.Error as err:
            if not err.matches(Gtk.DialogError.quark(), Gtk.DialogError.DISMISSED):
                logging.exception("Could not open files")
        else:
            safe_entry: SafeEntry = self.props.safe_entry

            def callback(gfile, result):
                try:
                    gbytes, _stream = gfile.load_bytes_finish(result)
                except GLib.Error:
                    logging.exception("Could not read attachment")
                else:
                    filename = gfile.get_basename()
                    data = gbytes.get_data()
                    new_attachment = safe_entry.add_attachment(data, filename)
                    self.add_attachment_row(new_attachment)

            for attachment in files:
                attachment.load_bytes_async(None, callback)

    def add_attachment_row(self, attachment: Attachment) -> None:
        safe_entry: SafeEntry = self.props.safe_entry
        attachment_row = AttachmentEntryRow(
            self.unlocked_database, safe_entry, attachment, self.attachment_list_box
        )
        self.attachment_list_box.append(attachment_row)

    @Gtk.Template.Callback()
    def _on_url_copy_button_clicked(self, _button):
        url = self.url_entry_row.props.text
        self.unlocked_database.send_to_clipboard(url, _("Address copied"))

    def show_row(self, row: Gtk.ListBoxRow, non_empty: bool, add_all: bool) -> None:
        if non_empty or add_all:
            row.set_visible(True)
        else:
            row.set_visible(False)

    def _on_safe_entry_updated(self, _safe_entry: SafeEntry) -> None:
        self.unlocked_database.start_database_lock_timer()

    def otp_update(self, safe_entry, _value):
        otp_token = safe_entry.otp_token()

        if self.otp_timer_handler is not None:
            GLib.source_remove(self.otp_timer_handler)
            self.otp_timer_handler = None

        if otp_token:
            remaining_time = safe_entry.otp_lifespan() / safe_entry.otp_interval()
            self.otp_progress_icon.props.progress = remaining_time
            self.otp_token_row.props.visible = True
            self.otp_token_row.props.title = otp_token
            self.otp_timer_handler = GObject.timeout_add(
                100,
                self.otp_update,
                safe_entry,
                None,
            )
        else:
            self.otp_token_row.props.visible = False

        if not otp_token and safe_entry.props.otp:
            self.otp_error_revealer.reveal(True)
            self.otp_secret_entry_row.add_css_class("error")
        else:
            self.otp_error_revealer.reveal(False)
            self.otp_secret_entry_row.remove_css_class("error")

        return GLib.SOURCE_CONTINUE



# SPDX-License-Identifier: GPL-3.0-only
from gi.repository import Gtk


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/entry_page_icon.ui")
class EntryPageIcon(Gtk.FlowBoxChild):
    __gtype_name__ = "EntryPageIcon"

    image = Gtk.Template.Child()

    def __init__(self, icon_name, icon_number):
        super().__init__()

        self.image.props.icon_name = icon_name
        self.props.tooltip_text = icon_name
        self.set_name(icon_number)



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from gettext import gettext as _

from gi.repository import Adw, GLib, GObject, Gtk

from gsecrets.safe_element import SafeEntry
from gsecrets.utils import format_time


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/expiration_date_row.ui")
class ExpirationDateRow(Adw.Bin):
    """Expiration date widget.

    All the dates are stored and gathered in UTC and showed in the local
    timezone.
    """

    __gtype_name__ = "ExpirationDateRow"

    action_row = Gtk.Template.Child()
    calendar = Gtk.Template.Child()
    popover = Gtk.Template.Child()
    remove_button = Gtk.Template.Child()
    popover_edit_button = Gtk.Template.Child()
    list_box = Gtk.Template.Child()

    _safe_entry = None

    def do_mnemonic_activate(self, group_cycling: bool) -> bool:
        # pylint: disable=arguments-differ
        return self.action_row.mnemonic_activate(group_cycling)

    @GObject.Property(type=SafeEntry)
    def safe_entry(self) -> SafeEntry:
        return self._safe_entry

    @safe_entry.setter  # type: ignore
    def safe_entry(self, entry: SafeEntry) -> None:
        self._safe_entry = entry

        entry.bind_property(
            "expires",
            self.remove_button,
            "sensitive",
            GObject.BindingFlags.SYNC_CREATE,
        )
        entry.connect("notify::expired", self.on_safe_entry_notify_expired)
        expiry_date = entry.expiry_time  # UTC

        if entry.props.expires and expiry_date:
            self.calendar.select_day(expiry_date.to_local())
            self.action_row.props.title = format_time(expiry_date, hours=False).lstrip()
            if entry.props.expired:
                self.action_row.props.subtitle = _("Entry expired")

    def on_safe_entry_notify_expired(
        self,
        safe_entry: SafeEntry,
        _gparam: GObject.ParamSpecBoolean,
    ) -> None:
        if safe_entry.props.expired:
            self.action_row.props.subtitle = _("Entry expired")
        else:
            self.action_row.props.subtitle = None

    @Gtk.Template.Callback()
    def on_edit_button_clicked(self, _button: Gtk.Button) -> None:
        safe_entry = self.props.safe_entry
        date = self.calendar.get_date().to_utc()
        safe_entry.expiry_time = date
        self.action_row.props.title = format_time(date, hours=False).lstrip()

        safe_entry.props.expires = True

        self.popover.popdown()

    @Gtk.Template.Callback()
    def on_remove_button_clicked(self, _button: Gtk.Button) -> None:
        safe_entry = self.props.safe_entry
        self.action_row.props.title = _("Expiration Date not Set")
        self.action_row.props.subtitle = None
        now = GLib.DateTime.new_now_utc()
        self.calendar.select_day(now)
        safe_entry.props.expires = False



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from gi.repository import Adw, Gio, GLib, GObject

from gsecrets.provider.file_provider import FileProvider
from gsecrets.provider.pkcs11_provider import Pkcs11Provider
from gsecrets.provider.yubikey_provider import YubiKeyProvider

if TYPE_CHECKING:
    from gsecrets.utils import LazyValue

KEY_PROVIDERS = [FileProvider, Pkcs11Provider, YubiKeyProvider]


class Providers(GObject.Object):
    def __init__(self, window: Adw.ApplicationWindow):
        super().__init__()

        self.providers = []
        self.salt: LazyValue[bytes] | None = None

        for key_provider in KEY_PROVIDERS:
            self.providers.append(key_provider(window))

    def get_key_providers(self) -> list:
        return self.providers

    def generate_composite_key_async(
        self,
        salt: LazyValue[bytes],
        callback: Gio.AsyncReadyCallback,
        cancellable: GLib.Cancellable = None,
    ) -> None:
        """Generate composite key.

        Traverse all available key providers and request each raw_key.
        This one is appended to the main raw_key.
        """
        self.salt = salt

        def generate_composite_key_task(task, self, _task_data, _cancellable):
            raw_key = b""

            # Generate composite key
            # It is safe to travel all providers as non configured ones do not
            # have a key.
            for provider in self.providers:
                logging.debug("Generate key for %s", type(provider).__name__)
                try:
                    if provider.generate_key(self.salt):
                        logging.debug("Adding key from %s", type(provider).__name__)
                        raw_key += provider.key
                except ValueError as err:
                    task.return_error(GLib.Error(str(err)))
                    return

            if len(raw_key) == 0:
                logging.debug("No key providers in use, returning None")
                task.return_value(None)
                return

            task.return_value(raw_key)

        task = Gio.Task.new(self, cancellable, callback)
        task.run_in_thread(generate_composite_key_task)

    def generate_composite_key_finish(self, result):
        """Finish generate_composite_key_finish.

        Can raise GLib.Error
        """
        _success, composite_key = result.propagate_value()
        return composite_key



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import secrets
import string
import threading
import typing

from gi.repository import GLib
from zxcvbn_rs_py import zxcvbn

if typing.TYPE_CHECKING:
    from collections.abc import Callable


def _satisfies_requirements(
    password: str,
    use_uppercase: bool,
    use_lowercase: bool,
    use_numbers: bool,
    use_symbols: bool,
) -> bool:
    if use_uppercase and not any(c.isupper() for c in password):
        return False

    if use_lowercase and not any(c.islower() for c in password):
        return False

    if use_numbers and not any(c.isdigit() for c in password):
        return False

    if use_symbols and not any(not c.isalnum() for c in password):  # noqa: SIM103
        return False

    return True


def generate(
    length: int,
    use_uppercase: bool,
    use_lowercase: bool,
    use_numbers: bool,
    use_symbols: bool,
) -> str:
    """Generate a password based on some criteria.

    :param int digits: password number of characters
    :param bool high_letter: password must contain uppercase letters
    :param bool low_letter: password must contain low letters
    :param bool numbers: password must contain digits
    :param bool special: password must contain special characters
    :returns: a password
    :rtype: str
    """
    characters: str = ""

    if use_uppercase:
        characters += string.ascii_uppercase

    if use_lowercase:
        characters += string.ascii_lowercase

    if use_numbers:
        characters += string.digits

    if use_symbols:
        characters += string.punctuation

    # If all options are disabled, generate a password using
    # arbitrary ASCII characters.
    # TODO Revisit this. It might be a sane default, but
    # it is highly un-intuitive.
    if characters == "":
        characters = (
            string.ascii_uppercase
            + string.ascii_lowercase
            + string.digits
            + string.punctuation
        )

    while True:
        password = "".join([secrets.choice(characters) for _ in range(length)])
        if _satisfies_requirements(
            password,
            use_uppercase,
            use_lowercase,
            use_numbers,
            use_symbols,
        ):
            return password


def strength(password: str) -> int:
    """Get strength of a password between 0 and 4.

    The higher the score is, the more secure the password is.

    :param str password: password to test
    :returns: strength of password
    :rtype: int
    """
    if password:
        return int(zxcvbn(password).score)

    return 0


def _threaded_compute_strength(password: str, callback: Callable[[int], None]) -> None:
    score = strength(password)

    def wrapper_callback(score):
        callback(score)
        return GLib.SOURCE_REMOVE

    GLib.idle_add(wrapper_callback, score)


def strength_async(password: str, callback: Callable[[int], None]) -> None:
    if password:
        thread = threading.Thread(
            target=_threaded_compute_strength,
            args=[password, callback],
        )
        thread.daemon = True
        thread.start()
    else:
        callback(0)



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import typing
from gettext import gettext as _

from gi.repository import Adw, Gtk

if typing.TYPE_CHECKING:
    from gsecrets.safe_element import SafeEntry


class ProtectedAttributeEntryRow:
    def __init__(self, entry: SafeEntry, key: str, value: str) -> None:
        super().__init__()

        self._entry = entry
        self._key = key

        self.row = Adw.PasswordEntryRow.new()

        self.row.props.title = key
        if value:
            self.row.props.text = value

        copy_btn = Gtk.Button.new_from_icon_name("edit-copy-symbolic")
        copy_btn.props.valign = Gtk.Align.CENTER
        copy_btn.props.tooltip_text = _("Copy Attribute")
        copy_btn.add_css_class("flat")
        copy_btn.connect("clicked", self._on_copy_button_clicked)

        button = Gtk.Button.new_from_icon_name("user-trash-symbolic")
        button.props.valign = Gtk.Align.CENTER
        button.props.tooltip_text = _("Remove Attribute")
        button.add_css_class("flat")
        button.connect("clicked", self._on_remove_button_clicked)

        self.row.connect("changed", self._on_changed)
        self.row.add_suffix(copy_btn)
        self.row.add_suffix(button)

    def _on_remove_button_clicked(self, _button):
        window = self.row.get_root()
        self._entry.delete_attribute(self._key)
        window.unlocked_db.attribute_deleted(
            self._entry,
            self._key,
            self.row.props.text,
            True,
        )

    def _on_changed(self, row):
        self._entry.set_attribute(self._key, row.props.text, True)

    def _on_copy_button_clicked(self, _row):
        clipboard = self.row.get_clipboard()
        clipboard.set(self.row.props.text)

        window = self.row.get_root()
        window.send_notification(_("Attribute copied"))



# SPDX-License-Identifier: GPL-3.0-only
import logging
from collections import deque

from gi.repository import Gio, GLib, GObject

from gsecrets.const import APP_ID

MAX_RECENT_ITEMS = 10


class RecentManager(GObject.Object):
    changed = GObject.Signal(arg_types=())

    def __init__(self):
        super().__init__()

        self.settings = Gio.Settings.new(APP_ID)

        self._load_items()
        self.settings.connect("changed::last-opened-list", self._on_settings_changed)

    def __iter__(self):
        return iter(self.recents)

    def __reversed__(self):
        return reversed(self.recents)

    def add_item(self, item: Gio.File) -> None:
        if found := next(
            (f for f in self.recents if f.get_uri() == item.get_uri()), None
        ):
            if bool(self.recents) and found != self.recents[-1]:
                self.recents.remove(found)
            else:
                return

        self.recents.append(item)
        self._save_items()
        self.emit(self.changed)

    def _remove_items(self, items: list[Gio.File]) -> None:
        for item in items:
            self.recents.remove(item)

        if bool(items):
            self._save_items()
            self.emit(self.changed)

    def is_empty(self) -> bool:
        return not bool(self.recents)

    def purge_items(self) -> None:
        if self.is_empty():
            return

        self.recents.clear()
        self._save_items()
        self.emit(self.changed)

    def _save_items(self):
        uris = [item.get_uri() for item in self.recents]
        self.settings.set_strv("last-opened-list", uris)

    def _load_items(self):
        uris = self.settings.get_strv("last-opened-list")
        items = [Gio.File.new_for_uri(uri) for uri in uris]
        self.recents = deque(items, MAX_RECENT_ITEMS)

    def _on_settings_changed(self, _settings, _key):
        self._load_items()
        self.emit(self.changed)

    async def clean_non_existent(self) -> None:
        to_remove = []

        for item in self.recents:
            try:
                await item.query_info_async(
                    Gio.FILE_ATTRIBUTE_STANDARD_TYPE,
                    Gio.FileQueryInfoFlags.NONE,
                    GLib.PRIORITY_DEFAULT,
                    None,
                )
            except GLib.Error as err:
                file_exist = not err.matches(
                    Gio.io_error_quark(), Gio.IOErrorEnum.NOT_FOUND
                )
                if file_exist:
                    logging.exception("Could not get file info")
                else:
                    uri = item.get_uri()
                    logging.info("Ignoring nonexistent recent file: %s", uri)

                to_remove.append(item)

        # We remove items after we are done iterating to avoid undefined
        # behaviour.
        self._remove_items(to_remove)



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from gettext import gettext as _

from gi.repository import Adw, GObject, Gtk, GtkSource


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/notes_dialog.ui")
class NotesDialog(Adw.Dialog):
    # pylint: disable=too-many-instance-attributes

    __gtype_name__ = "NotesDialog"

    _search_bar = Gtk.Template.Child()
    _search_button = Gtk.Template.Child()
    _toast_overlay = Gtk.Template.Child()
    _value_entry = Gtk.Template.Child()

    clipboard_timer_handler: None | int = None

    def __init__(self, unlocked_database, safe_entry):
        super().__init__()

        self.__unlocked_database = unlocked_database
        self.__safe_entry = safe_entry
        self.__search_stopped = False
        self.__signal_handle = 0

        self.__notes_buffer = self._value_entry.get_buffer()

        self.__tag = self.__notes_buffer.create_tag(None)

        manager = Adw.StyleManager.get_default()

        self.__setup_signals()

        self.__update_tag()
        self.__update_style_scheme(manager)

    def __setup_signals(self):
        manager = Adw.StyleManager.get_default()

        handle = self.__unlocked_database.database_manager.connect(
            "notify::locked",
            self.__on_locked,
        )
        self.__signal_handle = handle
        # Bind text to the corresponding safe entry.
        self.__safe_entry.bind_property(
            "notes",
            self.__notes_buffer,
            "text",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )

        self.__notes_buffer.connect(
            "notify::style-scheme", self.__on_style_scheme_notify
        )
        manager.connect("notify::dark", self.__on_dark_notify)

    def __on_style_scheme_notify(self, _buffer, _pspec):
        self.__update_tag()

    def __on_dark_notify(self, manager, _pspec):
        self.__update_style_scheme(manager)

    def __update_style_scheme(self, manager):
        if manager.props.dark:
            scheme_name = "Adwaita-dark"
        else:
            scheme_name = "Adwaita"

        scheme_manager = GtkSource.StyleSchemeManager.get_default()
        scheme = scheme_manager.get_scheme(scheme_name)
        self.__notes_buffer.set_style_scheme(scheme)

    def __update_tag(self):
        table = self.__notes_buffer.get_tag_table()

        if current_tag := self.__tag:
            table.remove(current_tag)

        self.__tag = self.__notes_buffer.create_tag(None)

        if scheme := self.__notes_buffer.get_style_scheme():  # noqa: SIM102
            if style := scheme.get_style("search-match"):
                style.apply(self.__tag)

                if rgba := self.__tag.props.background_rgba:
                    rgba.alpha = 1.0
                    current_tag.props.background_rgba = rgba

        self.__tag.set_priority(table.get_size() - 1)

    #
    # Events
    #

    @Gtk.Template.Callback()
    def _on_copy_button_clicked(self, _button):
        notes_buffer = self.__notes_buffer
        bounds = notes_buffer.get_selection_bounds()
        if bounds:
            message = _("Selection copied")
            start, end = bounds[0], bounds[1]
        else:
            message = ""
            start = notes_buffer.get_start_iter()
            end = notes_buffer.get_end_iter()

        self.__unlocked_database.send_to_clipboard(
            notes_buffer.get_text(start, end, False),
            message=message,
            toast_overlay=self._toast_overlay,
        )

    @Gtk.Template.Callback()
    def _on_search_button_toggled(self, _button):
        self.__unlocked_database.start_database_lock_timer()

        if self.__search_stopped is True:
            return

        self.__toggle_search_bar()

    def __toggle_search_bar(self):
        if self._search_bar.get_search_mode() is True:
            self._search_bar.set_search_mode(False)
        else:
            self._search_bar.set_search_mode(True)

    @Gtk.Template.Callback()
    def _on_search_entry_changed(self, entry):
        self.__unlocked_database.start_database_lock_timer()

        notes_buffer = self.__notes_buffer

        notes_buffer.remove_all_tags(
            notes_buffer.get_start_iter(),
            notes_buffer.get_end_iter(),
        )

        start = notes_buffer.get_start_iter()
        if entry.get_text():
            self.__do_search(notes_buffer, entry.get_text(), start)

    @Gtk.Template.Callback()
    def _on_search_stopped(self, _entry):
        self.__search_stopped = True
        self._search_button.set_active(False)
        self.__search_stopped = False

    #
    # Tools
    #

    def __do_search(self, notes_buffer, keyword, start):
        end = notes_buffer.get_end_iter()
        match = start.forward_search(keyword, Gtk.TextSearchFlags.CASE_INSENSITIVE, end)

        if match is not None:
            match_start, match_end = match
            notes_buffer.apply_tag(self.__tag, match_start, match_end)
            self.__do_search(notes_buffer, keyword, match_end)

    def __on_locked(self, database_manager, _value):
        locked = database_manager.props.locked
        if locked:
            self.close()

    def do_closed(self) -> None:
        if handle := self.__signal_handle:
            self.__unlocked_database.database_manager.disconnect(handle)
            self.__signal_handle = 0



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import typing
from gettext import gettext as _

from gi.repository import Adw, GObject, Gtk

from gsecrets.safe_element import SafeEntry

if typing.TYPE_CHECKING:
    from gsecrets.database_manager import DatabaseManager
    from gsecrets.unlocked_database import UnlockedDatabase


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/credentials_group.ui")
class CredentialsGroup(Adw.PreferencesGroup):
    """Widget to set the username and password of an entry."""

    __gtype_name__ = "CredentialsGroup"

    _copy_password_button = Gtk.Template.Child()
    _generate_password_button = Gtk.Template.Child()
    _password_entry_row = Gtk.Template.Child()
    _username_entry_row = Gtk.Template.Child()

    _unlocked_database: UnlockedDatabase | None = None

    _safe_entry = None

    @property
    def username(self):
        return self._username_entry_row.props.text

    @GObject.Property(type=SafeEntry)
    def entry(self):
        return self._safe_entry

    @entry.setter  # type: ignore
    def entry(self, entry):
        if self._safe_entry == entry:
            return

        self._safe_entry = entry
        self._password_entry_row.props.text = entry.props.password

        self._password_entry_row.get_delegate().connect(
            "notify::visibility",
            self._on_password_visibility_changed,
        )

        entry.bind_property(
            "username",
            self._username_entry_row,
            "text",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )
        entry.bind_property(
            "password",
            self._password_entry_row,
            "text",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )

    @property
    def unlocked_database(self) -> UnlockedDatabase | None:
        return self._unlocked_database

    @unlocked_database.setter  # type: ignore
    def unlocked_database(self, unlocked_database: UnlockedDatabase) -> None:
        self._db_manager: DatabaseManager = unlocked_database.database_manager
        self._unlocked_database = unlocked_database

    @Gtk.Template.Callback()
    def _on_copy_password_button_clicked(self, _widget: Gtk.Button) -> None:
        self.copy_password()

    @Gtk.Template.Callback()
    def on_password_generated(self, _popover, password):
        self._password_entry_row.props.text = password

    @Gtk.Template.Callback()
    def _on_password_value_changed(self, _entry: Adw.PasswordEntry) -> None:
        """Invoked when the password entry has changed.

        Take note that this callback is already invoked after the initial
        population of the entry when nothing has really been changed by
        the user, so be careful of doing things here that could have unwanted
        side-effects.
        """
        if self._unlocked_database:
            self._unlocked_database.start_database_lock_timer()

    @Gtk.Template.Callback()
    def _on_username_copy_button_clicked(self, _button):
        if self._unlocked_database:
            username = self._username_entry_row.props.text
            self._unlocked_database.send_to_clipboard(username, _("Username copied"))

    @Gtk.Template.Callback()
    def _on_apply(self, _entry_row):
        if u_db := self._unlocked_database:
            u_db.window.send_notification(_("Saved in history"))

        self._safe_entry.save_history()

    def copy_password(self) -> None:
        if self._unlocked_database:
            password: str = self._password_entry_row.props.text
            self._unlocked_database.send_to_clipboard(password, _("Password copied"))

    def _on_password_visibility_changed(self, _widget, _value):
        if self._password_entry_row.get_delegate().get_visibility():
            self._password_entry_row.add_css_class("monospace")
        else:
            self._password_entry_row.remove_css_class("monospace")



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import typing
from gettext import gettext as _

from gi.repository import Adw, GLib, GObject, Gtk

from gsecrets.safe_element import SafeGroup

if typing.TYPE_CHECKING:
    from gsecrets.unlocked_database import UnlockedDatabase


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/group_row.ui")
class GroupRow(Adw.Bin):
    __gtype_name__ = "GroupRow"

    _prefix_stack = Gtk.Template.Child()
    _selection_checkbox = Gtk.Template.Child()
    _group_icon = Gtk.Template.Child()

    _safe_group = None

    title = GObject.Property(type=str, default="")

    def __init__(self, unlocked_database):
        super().__init__()

        self._signals = GObject.SignalGroup.new(SafeGroup)
        self._bindings = GObject.BindingGroup.new()

        self._signals.connect_closure(
            "notify::name",
            self._on_group_name_changed,
            False,
        )
        self._signals.connect_closure(
            "notify::selected",
            self._on_selected_notify,
            False,
        )
        self._bindings.bind(
            "selected",
            self._selection_checkbox,
            "active",
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE,
        )
        self._bindings.bind(
            "sensitive",
            self,
            "sensitive",
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE,
        )

        self.unlocked_database = unlocked_database

        unlocked_database.connect(
            "notify::selection-mode",
            self._on_selection_mode_notify,
        )

    @GObject.Property(type=SafeGroup)
    def safe_group(self):
        return self._safe_group

    @safe_group.setter  # type: ignore
    def safe_group(self, element):
        if not isinstance(element, SafeGroup):
            msg = "Expected a SafeGroup."
            raise TypeError(msg)

        if element.is_trash_bin:
            self._group_icon.set_from_icon_name("user-trash-symbolic")
        else:
            self._group_icon.set_from_icon_name("folder-symbolic")

        self._safe_group = element

        self._signals.props.target = element
        self._bindings.props.source = element

        self._on_group_name_changed(element, None)
        self._on_selection_mode_notify(self.unlocked_database, None)

    @Gtk.Template.Callback()
    def _on_group_row_button_pressed(
        self,
        gesture: Gtk.GestureClick,
        _n_press: int,
        _event_x: float,
        _event_y: float,
    ) -> None:
        # pylint: disable=too-many-arguments
        gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        db_view: UnlockedDatabase = self.unlocked_database
        db_view.start_database_lock_timer()

        if db_view.props.selection_mode:
            selected = self._safe_group.selected  # type: ignore
            self._safe_group.selected = not selected  # type: ignore
        else:
            db_view.props.selection_mode = True
            self._safe_group.selected = True  # type: ignore

    def _on_selected_notify(self, group, _pspec):
        self.unlocked_database.start_database_lock_timer()

        if group.props.selected:
            self.unlocked_database.add_selection(group)
        else:
            self.unlocked_database.remove_selection(group)

    @Gtk.Template.Callback()
    def on_edit_button_clicked(self, _button: Gtk.Button) -> None:
        """Edit the button in a GroupRow when clicked.

        button: The edit button in the GroupRow
        """
        self.unlocked_database.start_database_lock_timer()  # Reset the lock timer

        self.unlocked_database.browsing_panel.selection_model.selected_item = (
            self._safe_group
        )

    def _on_group_name_changed(
        self,
        safe_group: SafeGroup,
        _value: GObject.ParamSpec,
    ) -> None:
        group_name = GLib.markup_escape_text(safe_group.name)
        if group_name:
            self.remove_css_class("italic-title")
            self.props.title = group_name
        else:
            self.add_css_class("italic-title")
            self.props.title = _("Title not Specified")

    def _on_selection_mode_notify(self, unlocked_db, _pspec):
        selection_mode = unlocked_db.props.selection_mode

        if selection_mode:
            visible_child = self._selection_checkbox
        else:
            visible_child = self._group_icon

        self._prefix_stack.props.visible_child = visible_child

    @Gtk.Template.Callback()
    def _on_long_press_gesture_pressed(self, gesture, _x, _y):
        gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        self.unlocked_database.props.selection_mode = True
        self._safe_group.props.selected = not self._safe_group.props.selected



# SPDX-License-Identifier: GPL-3.0-only
import math

from gi.repository import Gdk, GObject, Graphene, Gtk


class ProgressIcon(GObject.Object, Gdk.Paintable, Gtk.SymbolicPaintable):
    __gtype_name__ = "ProgressIcon"

    _progress: float = 0.0

    def do_snapshot_symbolic(self, snapshot, width, height, colors, _n_colors):
        color = colors[Gtk.SymbolicColor.FOREGROUND]
        color.alpha = 0.15

        rect = Graphene.Rect().alloc()
        rect.init(0, 0, width, height)

        ctx = snapshot.append_cairo(rect)
        ctx.set_source_rgba(color.red, color.green, color.blue, color.alpha)

        ctx.arc(width / 2, height / 2, width / 2, 0.0, 2 * math.pi)
        ctx.fill()

        if self.props.progress > 0.0:
            color.alpha = 1.0
            ctx.set_source_rgba(color.red, color.green, color.blue, color.alpha)
            ctx.arc(
                width / 2,
                height / 2,
                width / 2,
                -0.5 * math.pi,
                2 * self.props.progress * math.pi - 0.5 * math.pi,
            )
            if self.props.progress != 1.0:
                ctx.line_to(width / 2, height / 2)
                ctx.line_to(width / 2, 0)

            ctx.fill()

    @GObject.Property(type=float, default=0.0)
    def progress(self) -> float:
        return self._progress

    @progress.setter  # type: ignore
    def progress(self, progress: float) -> None:
        progress = max(min(progress, 1.0), 0.0)
        if self._progress != progress:
            self._progress = progress
            self.invalidate_contents()



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import logging
import typing
from gettext import gettext as _
from pathlib import Path

from gi.repository import Adw, Gio, GLib, Gtk

import gsecrets.config_manager
from gsecrets import const
from gsecrets.database_manager import DatabaseManager
from gsecrets.err import QUARK, ErrorType
from gsecrets.fingerprint_verifier import FingerprintVerifier
from gsecrets.unlocked_database import UnlockedDatabase
from gsecrets.utils import compare_passwords

if typing.TYPE_CHECKING:
    from gsecrets.widgets.window import Window


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/unlock_database.ui")
class UnlockDatabase(Adw.Bin):
    # pylint: disable=too-many-instance-attributes

    __gtype_name__ = "UnlockDatabase"

    database_manager: DatabaseManager | None = None

    password_entry = Gtk.Template.Child()
    key_group = Gtk.Template.Child()
    provider_group = Gtk.Template.Child()
    status_page = Gtk.Template.Child()
    headerbar = Gtk.Template.Child()
    unlock_button = Gtk.Template.Child()
    banner = Gtk.Template.Child()
    _progress = Gtk.Template.Child()
    fingerprint_img = Gtk.Template.Child()

    FPRINT_MAX_TRIES = 3

    def __init__(self, window: Window, database_file: Gio.File) -> None:
        super().__init__()

        filepath = database_file.get_path()

        self.window = window
        self.composition_key = None

        # Reset headerbar to initial state if it already exists.
        self.headerbar.title.props.title = database_file.get_basename()

        if database := self.window.unlocked_db:
            is_current = database.database_manager.path == filepath
            if is_current:
                self.database_manager = database.database_manager

        if not self.database_manager:
            self.database_manager = DatabaseManager(
                window.key_providers.get_key_providers(),
                filepath,
            )

        if gsecrets.const.IS_DEVEL:
            self.status_page.props.icon_name = gsecrets.const.APP_ID

        self.fprint: FingerprintVerifier | None = None
        self.fprint_tries = 0
        if gsecrets.config_manager.get_fingerprint_quick_unlock():
            self._map_fingerprint_reader()

        self.quick_unlock_failed = False

        self.event_controller = Gtk.EventControllerFocus.new()
        self.add_controller(self.event_controller)
        self.event_controller.connect("enter", self.on_enter)
        self.event_controller.connect("leave", self.on_leave)

        settings = self.window.application.settings
        settings.connect(
            "changed::fingerprint-quick-unlock",
            self._on_fingerprint_settings_changed,
        )

        self.window.set_default_widget(self.unlock_button)

        for key_provider in self.window.key_providers.get_key_providers():
            if key_provider.available:
                widget = key_provider.create_unlock_widget(self.database_manager)
                self.provider_group.add(widget)

    def do_unmap(self):  # pylint: disable=arguments-differ
        Gtk.Widget.do_unmap(self)
        self._progress.end_pulse()
        self.remove_controller(self.event_controller)
        self._unmap_fingerprint_reader()

    def on_enter(self, _user_data):
        if self.database_manager and self.database_manager.password != "":
            if (
                gsecrets.config_manager.get_quick_unlock()
                and not self.quick_unlock_failed
            ):
                self.window.show_banner(_("Quick Unlock active"))
                self.provider_group.props.visible = False
            # only start reading if there are still tries left
            if self.fprint_tries < self.FPRINT_MAX_TRIES:
                self._start_fingerprint_reader()
        else:
            logging.debug("Quick unlock disabled as no password is available.")

    def on_leave(self, _user_data):
        if self.fprint_tries < self.FPRINT_MAX_TRIES:
            self._stop_fingerprint_reader()

    def grab_entry_focus(self):
        self.password_entry.grab_focus()

    def is_safe_open_elsewhere(self) -> bool:
        """Whether the safe is already open.

        Return True if the safe is already open but not in the
        current window.
        """
        is_current = False
        db_path = self.database_manager.path  # type: ignore
        is_open = self.window.application.is_safe_open(db_path)

        if database := self.window.unlocked_db:
            is_current = database.database_manager.path == db_path

        return is_open and not is_current

    def _on_generated_composite_key(self, providers, result):
        self._set_sensitive(True)

        try:
            self.composition_key = providers.generate_composite_key_finish(result)
        except GLib.Error:
            logging.exception("Could not generate composite key")
            self.window.send_notification(_("Failed to generate composite key"))
            return

        entered_pwd = self.password_entry.get_text()

        if self.is_safe_open_elsewhere():
            self.window.send_notification(
                # TRANSLATORS For example: "Safe ~/MySafe.kdbx is already open"
                _("Safe {path} is already open").format(
                    path=self.database_manager.path,
                ),
            )
            return

        if not self.database_manager.opened:
            self._open_database()
            return

        quick_unlock = (
            gsecrets.config_manager.get_quick_unlock()
            and not self.quick_unlock_failed
            and entered_pwd == self.database_manager.password[-4:]
        )

        full_unlock = (
            compare_passwords(entered_pwd, self.database_manager.password)
            and self.database_manager.composition_key == self.composition_key
        )

        if quick_unlock or full_unlock:
            self.database_manager.props.locked = False
            self.database_manager.add_to_history()
        else:
            self._unlock_failed()

    @Gtk.Template.Callback()
    def _on_entry_activated(self, _widget: Gtk.Widget) -> None:
        self.unlock_button.activate()

    @Gtk.Template.Callback()
    def _on_unlock_button_clicked(self, _widget: Gtk.Button) -> None:
        if not self.database_manager:
            return

        self._set_sensitive(False)
        self.window.key_providers.generate_composite_key_async(
            self.database_manager.get_salt_as_lazy(),
            self._on_generated_composite_key,
        )

    #
    # Open Database
    #

    def _open_database(self):
        self._progress.start_pulse()

        self._set_sensitive(False)

        password = self.password_entry.props.text

        self.database_manager.unlock_async(
            password,
            self.composition_key,
            self._unlock_callback,
        )

    def _unlock_callback(self, database_manager, result):
        try:
            database_manager.unlock_finish(result)
        except GLib.Error as err:
            if not err.matches(QUARK, ErrorType.CREDENTIALS_ERROR):
                logging.exception("Could not unlock safe")

            self._unlock_failed()
            return

        opened = Gio.File.new_for_path(database_manager.path)

        self.window.action_set_enabled("win.new_database", False)

        if gsecrets.config_manager.get_development_backup_mode():
            self.store_backup(opened)

        database_manager.add_to_history()

        if self.window.unlocked_db is None:
            database = UnlockedDatabase(self.window, database_manager)
            self.window.unlocked_db = database
            self.window.unlocked_db_bin.props.child = database

        self.window.view = self.window.View.UNLOCKED_DATABASE
        self._reset_page()

    #
    # Helper Functions
    #

    def _unlock_failed(self) -> None:
        self.window.send_notification(_("Failed to unlock Safe"))

        self.password_entry.add_css_class("error")
        self.password_entry.delete_text(0, -1)
        self._set_sensitive(True)
        self._progress.end_pulse()

        self.quick_unlock_failed = True

        # Regrab the focus of the entry.
        self.password_entry.grab_focus()

    def _reset_page(self):
        self.password_entry.set_text("")
        self.password_entry.remove_css_class("error")

        self._set_sensitive(True)

        self._progress.end_pulse()

        self.quick_unlock_failed = True
        self.provider_group.props.visible = True

    def _set_sensitive(self, sensitive):
        delegate = self.password_entry.get_delegate()
        if delegate.has_focus() and not sensitive:
            self.window.props.focus_widget = None

        for key in self.key_group:
            key.set_sensitive(sensitive)

        for key in self.provider_group:
            key.set_sensitive(sensitive)

        self.unlock_button.set_sensitive(sensitive)
        self.headerbar.set_sensitive(sensitive)

    def store_backup(self, gfile):
        cache_dir = Path(GLib.get_user_cache_dir()) / const.SHORT_NAME / "backup"
        if not cache_dir.exists():
            cache_dir.mkdir(parents=True)

        current_time = GLib.DateTime.new_now_local().format("%F_%T")
        basename = Path(gfile.get_basename()).stem
        backup_name = basename + "_backup_" + current_time + ".kdbx"
        backup = Gio.File.new_for_path(str(cache_dir / backup_name))

        def callback(gfile, result):
            try:
                gfile.copy_finish(result)
            except GLib.Error as err:
                logging.warning("Could not save database backup: %s", err.message)

        if GLib.check_version(2, 81, 0) is None:
            gfile.copy_async(
                backup,
                Gio.FileCopyFlags.NONE,
                GLib.PRIORITY_DEFAULT,
                None,
                None,
                callback,
            )
        else:
            gfile.copy_async(
                backup,
                Gio.FileCopyFlags.NONE,
                GLib.PRIORITY_DEFAULT,
                None,
                None,
                None,
                callback,
            )

    def show_banner(self, label: str) -> None:
        self.banner.set_title(label)
        self.banner.set_revealed(True)

    def close_banner(self):
        self.banner.set_revealed(False)

    #
    # Fingerprint-related stuff
    #

    def _on_fingerprint_settings_changed(self, _settings, _key):
        if gsecrets.config_manager.get_fingerprint_quick_unlock():
            logging.debug("Fingerprint got enabled, mapping fingerprint reader...")
            self._map_fingerprint_reader()
        else:
            logging.debug("Fingerprint got disabled, unmapping fingerprint reader...")
            self._unmap_fingerprint_reader()

    def _start_fingerprint_reader_cb(self) -> None:
        if not self.fprint:
            return

        async def cb():
            if await self.fprint.verify_start():
                self.fingerprint_img.props.visible = True

        app = Gio.Application.get_default()
        app.create_asyncio_task(cb())

    def _start_fingerprint_reader(self) -> None:
        if not self.fprint:
            return
        self.fprint.connect(self._start_fingerprint_reader_cb)

    def _stop_fingerprint_reader(self) -> None:
        if not self.fprint:
            return

        async def disconnect():
            if fprint := self.fprint:
                await fprint.verify_stop()
                fprint.disconnect()

            self.fingerprint_img.props.visible = False

        app = Gio.Application.get_default()
        app.create_asyncio_task(disconnect())

    def _map_fingerprint_reader(self) -> None:
        logging.debug("Connecting to the fingerprint device...")
        try:
            self.fprint = FingerprintVerifier(
                self._on_fingerprint_success,
                self._on_fingerprint_retry,
                self._on_fingerprint_failure,
            )
        except RuntimeError as err:
            logging.debug("Failed initialize fingerprint: %s", err)

    def _unmap_fingerprint_reader(self) -> None:
        if not self.fprint:
            return
        logging.debug("Disconnecting the fingerprint device...")
        self._stop_fingerprint_reader()
        self.fprint = None

    def _on_fingerprint_success(self):
        """Success callback of the the FingerprintVerifier."""
        self.fingerprint_img.add_css_class("success")
        GLib.timeout_add(250, self._after_fingerprint_success)
        logging.debug("Fingerprint success")

    def _after_fingerprint_success(self):
        """Unlocks database after animation of _on_fingerprint_success."""
        self.database_manager.props.locked = False

    def _on_fingerprint_retry(self):
        """Retry callback of the the FingerprintVerifier."""
        self.fingerprint_img.add_css_class("retry")
        GLib.timeout_add(850, self._after_fingerprint_retry)
        logging.debug("Fingerprint retry")

    def _after_fingerprint_retry(self):
        """Cleans up animation after _on_fingerprint_retry."""
        self.fingerprint_img.remove_css_class("retry")

    def _on_fingerprint_failure(self):
        """Failure callback of the the FingerprintVerifier."""
        self.fprint_tries += 1

        async def start_fprint():
            status = await self.fprint.verify_start()
            if not status:
                # in case fingerprint sensor itself refuses
                self.fingerprint_img.add_css_class("error")
                self.window.send_notification(_("Please unlock using the password."))
                logging.debug(
                    "Retry after verify-no-match failed. Fingerprint disabled.",
                )

        if self.fprint_tries >= self.FPRINT_MAX_TRIES:
            # at most 3 tries for unlocking with the fingerprint sensor
            self.fingerprint_img.add_css_class("error")
            self.window.send_notification(
                _("Maximum tries reached. Please unlock using the password."),
            )
            if fprint := self.fprint:
                fprint.disconnect()

            logging.debug("Max tries reached. Fingerprint disabled.")
        else:
            # else start verification again
            self.fingerprint_img.add_css_class("warning")
            GLib.timeout_add(850, self._after_fingerprint_warning)
            logging.debug("Fingerprint no-match for try %i", self.fprint_tries)
            app = Gio.Application.get_default()
            app.create_asyncio_task(start_fprint())

    def _after_fingerprint_warning(self):
        """Cleans up.

        Animation after _on_fingerprint_failure if FPRINT_MAX_TRIES not reached
        """
        self.fingerprint_img.remove_css_class("warning")



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable
from typing import Any, cast

from gi.repository import Gio, GLib

BUS_NAME = "net.reactivated.Fprint"
DBUS_TIMEOUT = 500  # In milliseconds
VERIFY_STATUS_TUPLE = 2


class FingerprintVerifier:
    """Provides easy access to a fingerprint device."""

    def __init__(
        self,
        on_success: Callable,
        on_retry: Callable,
        on_failure: Callable,
    ) -> None:
        """Constructor.

        Might throw an exception if dbus is unavailable, fprintd is missing or
        no device is available.

        Arguments:
        ---------
            on_success: Function to be called on success.
            on_retry: Function to be called if the fingerprint sensor suggests retrying.
            on_failure: Function to be called if the fingerprint did not match or the
                        hardware refused in the process.

        """
        self.on_success = on_success
        self.on_retry = on_retry
        self.on_failure = on_failure

        self._is_claimed = False
        self._device_proxy = None
        self._signal = None

    def connect(self, cb: Gio.Callback) -> None:
        """Initializes the connection to the fingerprint device.

        Shouldn't be called manually unless disconnect was called on the object before.

        Might throw an exception if dbus is unavailable, fprindt is missing
        or no device is available.
        """
        if self._is_claimed:
            return

        async def connect_fingerprint():
            con = await Gio.bus_get(Gio.BusType.SYSTEM, None)

            manager_proxy = await Gio.DBusProxy.new(
                con,
                Gio.DBusProxyFlags.NONE,
                None,
                BUS_NAME,
                "/net/reactivated/Fprint/Manager",
                "net.reactivated.Fprint.Manager",
                None,
            )
            if manager_proxy is None:
                logging.debug("Fprintd not available.")
                return

            try:
                variant = await manager_proxy.call(
                    "GetDefaultDevice",
                    None,
                    Gio.DBusCallFlags.NO_AUTO_START,
                    DBUS_TIMEOUT,
                    None,
                )
                device_str = variant[0]
            except GLib.GError:
                logging.exception("Failed to get default fingerprint device")
                return

            if device_str == "":
                logging.debug("No fingerprint sensor found!")
                return

            self._device_proxy = await Gio.DBusProxy.new(
                con,
                Gio.DBusProxyFlags.NONE,
                None,
                BUS_NAME,
                device_str,
                "net.reactivated.Fprint.Device",
                None,
            )
            if self._device_proxy is None:
                logging.debug("Failed to find the default fingerprint device.")
                return

            self._signal = self._device_proxy.connect("g-signal", self._signal_handler)
            logging.debug("Successfully connected to the fingerprint device.")

            cb()

        app = Gio.Application.get_default()
        app.create_asyncio_task(connect_fingerprint())

    def disconnect(self) -> None:
        """Removes the handlers from the signal.

        Should be called when the class is not needed anymore.
        """
        if self._is_claimed or self._device_proxy is None:
            return

        self._is_claimed = False
        self._device_proxy.disconnect(self._signal)
        self._signal = None
        self._device_proxy = None
        logging.debug("Disconnected from the fingerprint device.")

    async def verify_start(self) -> bool:
        """Starts the fingerprint verification process.

        Returns: True if the fingerprint verification started.

        """
        if self._device_proxy is None:
            return False

        try:
            await self._claim_device()
        except GLib.Error:
            logging.exception("Failed to claim device")
            return False

        self._is_claimed = True
        try:
            await self._verify_start()
        except GLib.Error:
            logging.exception("Failed to start fingerprint verification on the device")
            return False

        logging.debug("Claimed fingerprint device, verification ongoing...")
        return True

    async def verify_stop(self) -> None:
        """Stops the fingerprint verification process."""
        if not self._is_claimed:
            return

        self._is_claimed = False

        try:
            await self._verify_stop()
            await self._release_device()
            logging.debug("Stopped verification, released fingerprint device.")
        except GLib.Error as err:
            logging.debug("Exception while releasing fingerprint device: %s", err)

    #
    # Internal methods
    #

    async def _claim_device(self) -> None:
        if self._device_proxy:
            await self._device_proxy.call(
                "Claim",
                GLib.Variant("(s)", ("",)),
                Gio.DBusCallFlags.NO_AUTO_START,
                DBUS_TIMEOUT,
                None,
            )

    async def _release_device(self) -> None:
        if self._device_proxy:
            await self._device_proxy.call(
                "Release",
                None,
                Gio.DBusCallFlags.NO_AUTO_START,
                DBUS_TIMEOUT,
                None,
            )

    async def _verify_start(self) -> None:
        if self._device_proxy:
            await self._device_proxy.call(
                "VerifyStart",
                GLib.Variant("(s)", ("any",)),
                Gio.DBusCallFlags.NO_AUTO_START,
                DBUS_TIMEOUT,
                None,
            )

    async def _verify_stop(self) -> None:
        if self._device_proxy:
            await self._device_proxy.call(
                "VerifyStop",
                None,
                Gio.DBusCallFlags.NO_AUTO_START,
                DBUS_TIMEOUT,
                None,
            )

    retry_results = [
        "verify-retry-scan",
        "verify-swipe-too-short",
        "verify-finger-not-centered",
        "verify-remove-and-retry",
    ]

    def _signal_handler(
        self,
        _proxy: Gio.DBusProxy,
        _sender: str,  # pylint: disable=unused-argument
        signal: str,
        args: tuple[Any],
    ) -> None:
        if signal != "VerifyStatus" or len(args) != VERIFY_STATUS_TUPLE:
            return

        result, done = cast(tuple[str, bool], args)  # see mypy/issues/1178

        if result in self.retry_results and not done:
            self.on_retry()

        if done:
            if result != "verify-disconnected":
                app = Gio.Application.get_default()
                app.create_asyncio_task(self.verify_stop())

            if result == "verify-match":
                self.on_success()
            else:
                self.on_failure()



# SPDX-License-Identifier: GPL-3.0-only
from gi.repository import Gio, GObject, Gtk

from gsecrets.safe_element import SafeElement


class SingleSelection(GObject.Object, Gio.ListModel, Gtk.SelectionModel):
    def __init__(self, model):
        super().__init__()

        self._inner = Gtk.SingleSelection.new()
        self._inner.props.autoselect = False
        self._inner.props.can_unselect = True
        self._inner.props.model = model

        self._item_pos = Gtk.INVALID_LIST_POSITION
        self._hide_selected = False
        self._selected_item = None

        self._inner.connect("items-changed", self._on_items_changed)

    def do_get_item(self, pos):
        return self._inner.get_item(pos)

    def do_get_n_items(self):
        return self._inner.get_n_items()

    def do_get_item_type(self):
        return self._inner.get_item_type()

    def do_is_selected(self, pos):
        item_pos = self._item_pos
        if self._hide_selected or item_pos == Gtk.INVALID_LIST_POSITION:
            return False

        return pos == item_pos

    @GObject.Property(
        type=bool,
        default=False,
        flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.EXPLICIT_NOTIFY,
    )
    def hide_selected(self):
        return self._hide_selected

    @hide_selected.setter  # type: ignore
    def hide_selected(self, hide_selected):
        if self._hide_selected == hide_selected:
            return

        self._hide_selected = hide_selected

        if self._item_pos != Gtk.INVALID_LIST_POSITION:
            self.selection_changed(self._item_pos, 1)

        self.notify("hide_selected")

    @GObject.Property(
        type=SafeElement,
        flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.EXPLICIT_NOTIFY,
    )
    def selected_item(self):
        return self._selected_item

    def selected_item_pos(self) -> int:
        return self._item_pos

    @selected_item.setter  # type: ignore
    def selected_item(self, selected_item):
        if self._selected_item == selected_item:
            return

        pos = self._find_item_position(selected_item, 0, self.get_n_items())
        self._set_selected_item_inner(selected_item, pos)

    def _set_selected_item_inner(self, selected_item, pos):
        if self._item_pos == pos:
            return

        self._selected_item = selected_item

        old_pos = self._item_pos
        self._item_pos = pos

        if not self._hide_selected:
            if old_pos == Gtk.INVALID_LIST_POSITION:
                self.selection_changed(pos, 1)
            elif pos == Gtk.INVALID_LIST_POSITION:
                self.selection_changed(old_pos, 1)
            elif pos < old_pos:
                self.selection_changed(pos, old_pos - pos + 1)
            else:
                self.selection_changed(old_pos, pos - old_pos + 1)

        self.notify("selected-item")

    def _find_item_position(self, item, start_pos, end_pos):
        model = self._inner.props.model
        for pos in range(start_pos, end_pos):
            if model.get_item(pos) == item:
                return pos

        return Gtk.INVALID_LIST_POSITION

    def _on_items_changed(self, _model, position, removed, added):
        item_position = self._item_pos

        if selected_item := self._selected_item:
            if item_position == Gtk.INVALID_LIST_POSITION:
                # Maybe the item got newly added
                self._item_pos = self._find_item_position(
                    selected_item,
                    position,
                    position + added,
                )
            elif item_position < position:
                # Nothing to do, position stays the same
                pass
            elif item_position < position + removed:
                self._item_pos = self._find_item_position(
                    selected_item,
                    position,
                    position + added,
                )
            elif item_position + added >= removed:
                self._item_pos = item_position + added - removed
            else:
                # This should never happen.
                self._item_pos = 0

        self.items_changed(position, removed, added)



# SPDX-License-Identifier: GPL-3.0-only
# Functions to sort a List Model with groups and entries.

from __future__ import annotations

import typing
from enum import IntEnum

from gi.repository import Gtk

from gsecrets.safe_element import SafeElement


class SortingHat:
    """Provides a variety of sorting algorithms."""

    class SortOrder(IntEnum):
        ASC = 0
        DEC = 1
        CTIME_ASC = 2
        CTIME_DEC = 3

    @staticmethod
    def get_sorter(order: SortOrder) -> typing.Callable:
        match order:
            case SortingHat.SortOrder.ASC:
                expr = Gtk.PropertyExpression.new(SafeElement.__gtype__, None, "name")
                return Gtk.StringSorter.new(expr)
            case SortingHat.SortOrder.DEC:
                expr = Gtk.PropertyExpression.new(SafeElement.__gtype__, None, "name")
                return Gtk.StringSorter.new(expr)
            case SortingHat.SortOrder.CTIME_ASC:
                expr = Gtk.PropertyExpression.new(
                    SafeElement.__gtype__, None, "ctime_int"
                )
                return Gtk.NumericSorter.new(expr)
            case SortingHat.SortOrder.CTIME_DEC:
                expr = Gtk.PropertyExpression.new(
                    SafeElement.__gtype__, None, "ctime_int"
                )
                sorter = Gtk.NumericSorter.new(expr)
                sorter.props.sort_order = Gtk.SortType.DESCENDING
                return sorter



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import typing
from gettext import gettext as _

from gi.repository import Adw, Gtk

if typing.TYPE_CHECKING:
    from gsecrets.safe_element import SafeEntry


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/attribute_entry_row.ui")
class AttributeEntryRow(Adw.EntryRow):
    __gtype_name__ = "AttributeEntryRow"

    def __init__(self, entry: SafeEntry, key: str, value: str) -> None:
        super().__init__()

        self.entry = entry
        self.key = key

        self.props.title = key
        if value:
            self.props.text = value

    @Gtk.Template.Callback()
    def _on_remove_button_clicked(self, _button):
        window = self.get_root()
        self.entry.delete_attribute(self.key)
        window.unlocked_db.attribute_deleted(
            self.entry,
            self.key,
            self.props.text,
            False,
        )

    @Gtk.Template.Callback()
    def _on_copy_button_clicked(self, _row):
        clipboard = self.get_clipboard()
        clipboard.set(self.props.text)
        window = self.get_root()
        window.send_notification(_("Attribute copied"))

    @Gtk.Template.Callback()
    def _on_changed(self, row):
        self.entry.set_attribute(self.key, row.get_text())



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from gettext import gettext as _

from gi.repository import Adw, GObject, Gtk

from gsecrets.password_generator import strength_async


class PasswordLevelBar(Adw.Bin):
    __gtype_name__ = "PasswordLevelBar"

    _password: str = ""

    def __init__(self) -> None:
        super().__init__()

        self.level_bar = Gtk.LevelBar()
        self.level_bar.update_property(
            [Gtk.AccessibleProperty.LABEL],
            [_("Password Strength")],
        )

        self.level_bar.props.max_value = 4.0
        self.level_bar.props.mode = Gtk.LevelBarMode.DISCRETE

        # Values in a discrete mode are rounded
        # instead of using a floor function.
        self.level_bar.add_offset_value("insecure", 0.49)
        self.level_bar.add_offset_value("weak", 1.49)
        self.level_bar.add_offset_value("medium", 2.49)
        self.level_bar.add_offset_value("strong", 3.49)
        self.level_bar.add_offset_value("secure", 4.0)

        self.set_child(self.level_bar)

    @GObject.Property(type=str, default="")
    def password(self) -> str:
        return self._password

    @password.setter  # type: ignore
    def password(self, password: str) -> None:
        self._password = password

        def on_password(strength):
            self.level_bar.props.value = strength

        strength_async(password, on_password)



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import typing
from gettext import gettext as _

from gi.repository import Adw, Gio, GObject

from gsecrets.safe_element import SafeElement

if typing.TYPE_CHECKING:
    from gsecrets.unlocked_database import UnlockedDatabase


class SelectionManager(GObject.Object):
    __gtype_name__ = "SelectionManager"

    entries_selected: list[SafeElement] = []
    groups_selected: list[SafeElement] = []

    entries_cut: list[SafeElement] = []
    groups_cut: list[SafeElement] = []

    hidden_rows = Gio.ListStore.new(SafeElement)

    selected_elements = GObject.Property(type=int, default=0)
    cut_mode = GObject.Property(type=bool, default=True)

    def __init__(self, unlocked_database):
        super().__init__()

        self.unlocked_database = unlocked_database

        unlocked_database.connect(
            "notify::selection-mode",
            self._on_selection_mode_changed,
        )

    def _on_selection_mode_changed(
        self,
        unlocked_database: UnlockedDatabase,
        _value: GObject.ParamSpec,
    ) -> None:
        if not unlocked_database.props.selection_mode:
            self._clear_all()

    # Events

    def delete_selection(self):
        # Abort the operation if there is a groups which is in the pathbar,
        # i.e. if it is a parent of the current view.
        for safe_group in self.groups_selected:
            if self.unlocked_database.database_manager.parent_checker(
                self.unlocked_database.current_element,
                safe_group.group,
            ):
                self.unlocked_database.window.send_notification(
                    _("Operation aborted: deleting currently active group"),
                )
                return

        # Before deleting we check if there is a mix of elements in the trash
        # bin and regular elements
        in_trash = False
        outside_trash = False
        for element in self.entries_selected:
            parent = element.parentgroup
            if parent.is_trash_bin:
                in_trash = True
            else:
                outside_trash = True

        for element in self.groups_selected:
            parent = element.parentgroup
            if parent.is_trash_bin or element.is_trash_bin:
                in_trash = True
            else:
                outside_trash = True

        mixed = in_trash and outside_trash

        def delete_elements():
            # List of possible undos
            undo_elements = []
            for element in self.entries_selected:
                parent = element.parentgroup
                if not element.trash():
                    self.unlocked_database.browsing_panel.unselect(element)
                    undo_elements.append((element, parent))

            for element in self.groups_selected:
                parent = element.parentgroup
                self.unlocked_database.browsing_panel.unselect(element)
                if not element.trash():
                    undo_elements.append((element, parent))

            self.unlocked_database.deleted_notification(undo_elements)
            self._clear_all()

        if mixed:

            def response_delete_cb(_dialog, _response):
                delete_elements()

            dialog = Adw.AlertDialog.new(
                _("Warning"),
                _(
                    "You are deleting elements in the trash bin, these deletions cannot be undone.",  # noqa: E501
                ),
            )
            dialog.add_response("cancel", _("_Cancel"))
            dialog.add_response("delete", _("_Delete"))
            dialog.set_response_appearance("delete", Adw.ResponseAppearance.DESTRUCTIVE)
            dialog.connect("response::delete", response_delete_cb)
            dialog.present(self.unlocked_database)
        else:
            delete_elements()

    def cut_selection(self):
        self.entries_cut = self.entries_selected
        self.groups_cut = self.groups_selected
        for group in self.groups_selected:
            group.props.sensitive = False
            self.hidden_rows.append(group)

        for entry in self.entries_selected:
            entry.props.sensitive = False
            self.hidden_rows.append(entry)

        self.props.cut_mode = False

    def paste_selection(self):
        # Abort the entire operation if one of the selected groups is a parent of
        # the current group.
        for safe_group in self.groups_cut:
            current_element = self.unlocked_database.current_element
            if self.unlocked_database.database_manager.parent_checker(
                current_element,
                safe_group.group,
            ):
                self.unlocked_database.window.send_notification(
                    _("Operation aborted: moving currently active group"),
                )
                self._clear_all()
                return

        current_element = self.unlocked_database.current_element

        for safe_entry in self.entries_cut:
            safe_entry.move_to(current_element)

        for safe_group in self.groups_cut:
            safe_group.move_to(current_element)

        self.unlocked_database.window.send_notification(_("Move completed"))
        self._clear_all()

    # Helpers

    def add_entry(self, entry: SafeElement) -> None:
        """Add an entry to selection.

        :param EntryRow group: entry_row to add
        """
        if entry not in self.entries_selected:
            self.entries_selected.append(entry)
            self._update_selection()

    def remove_entry(self, entry: SafeElement) -> None:
        """Remove an entry from selection.

        :param EntryRow group: entry_row to remove
        """
        if entry in self.entries_selected:
            self.entries_selected.remove(entry)

        self._update_selection()

    def add_group(self, group: SafeElement) -> None:
        """Add a group to selection.

        :param GroupRow group: group_row to add
        """
        if group not in self.groups_selected:
            self.groups_selected.append(group)
            self._update_selection()

    def remove_group(self, group: SafeElement) -> None:
        """Remove a group from selection.

        :param GroupRow group: group_row to remove
        """
        if group in self.groups_selected:
            self.groups_selected.remove(group)

        self._update_selection()

    def _update_selection(self) -> None:
        self.selected_elements = len(self.entries_selected) + len(self.groups_selected)

    def _clear_all(self) -> None:
        """Reset everything to the default state."""
        for group in reversed(self.groups_selected):
            group.selected = False

        for entry in reversed(self.entries_selected):
            entry.selected = False

        self.entries_selected.clear()
        self.groups_selected.clear()

        for element in self.hidden_rows:
            element.props.sensitive = True

        self.selected_elements = 0
        self.cut_mode = True
        self.entries_cut.clear()

        self.groups_cut.clear()
        self.hidden_rows.remove_all()

    def clear_selection(self):
        """Emit when the selection is cleared.

        Signal emitted to tell all list models that they should unselect
        their entries. It differs from the action app.selection.none, since
        the later removes selected entries only for the visible listbox.
        """
        for element in reversed(self.entries_selected):
            element.props.selected = False

        for element in reversed(self.groups_selected):
            element.props.selected = False



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import logging
import os
import typing
from gettext import gettext as _
from gettext import ngettext
from pathlib import Path

from gi.repository import Adw, Gio, GLib, GObject, Gtk

import gsecrets.config_manager
from gsecrets import const
from gsecrets.entry_page import EntryPage
from gsecrets.entry_row import EntryRow
from gsecrets.group_page import GroupPage
from gsecrets.group_row import GroupRow
from gsecrets.safe_element import SafeElement, SafeEntry, SafeGroup
from gsecrets.selection_manager import SelectionManager
from gsecrets.unlocked_headerbar import UnlockedHeaderBar
from gsecrets.widgets.browsing_panel import BrowsingPanel
from gsecrets.widgets.database_settings_dialog import DatabaseSettingsDialog
from gsecrets.widgets.properties_dialog import PropertiesDialog
from gsecrets.widgets.references_dialog import ReferencesDialog
from gsecrets.widgets.saving_conflict_dialog import SavingConflictDialog

if typing.TYPE_CHECKING:
    from gsecrets.database_manager import DatabaseManager
    from gsecrets.widgets.window import Window


class UndoData:
    def __init__(self, elements, toast):
        self.elements = elements
        self.toast = toast


class AttributeUndoData:
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        toast: Adw.Toast,
        entry: SafeEntry,
        key: str,
        value: str,
        protected: bool = False,
    ):
        self.toast = toast
        self.key = key
        self.value = value
        self.entry = entry
        self.protected = protected


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/unlocked_database.ui")
class UnlockedDatabase(Adw.BreakpointBin):
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-public-methods
    __gtype_name__ = "UnlockedDatabase"

    # Connection handlers
    db_locked_handler: int | None = None
    # Used only to delete the clipboard if the window is not active.
    on_is_active_handler: int | None = None
    _current_element: SafeElement | None = None
    _lock_timer_handler: int | None = None
    save_loop: int | None = None  # If int, a thread periodically saves the database
    session_handler_id: int | None = None

    headerbar_stack = Gtk.Template.Child()
    _split_view = Gtk.Template.Child()
    toolbar_view = Gtk.Template.Child()
    _select_entry_status_page = Gtk.Template.Child()
    _empty_page = Gtk.Template.Child()

    _search_active: bool = False
    _search_bar = Gtk.Template.Child()
    _search_entry = Gtk.Template.Child()

    _selection_mode: bool = False
    _selection_mode_headerbar = Gtk.Template.Child()
    _selection_mode_action_bar = Gtk.Template.Child()
    _delete_selection_button = Gtk.Template.Child()
    _cut_selection_button = Gtk.Template.Child()
    _paste_selection_button = Gtk.Template.Child()
    _selection_mode_title = Gtk.Template.Child()
    _cut_paste_button_stack = Gtk.Template.Child()
    _clear_selection_button = Gtk.Template.Child()

    undo_data: UndoData | None = None
    attribute_undo_data: AttributeUndoData | None = None

    def __init__(self, window: Window, dbm: DatabaseManager):
        super().__init__()
        # Instances
        self.window: Window = window
        self.database_manager: DatabaseManager = dbm

        root_group = SafeGroup.get_root(dbm)
        self.props.current_element = root_group

        # Headerbars
        self.headerbar = UnlockedHeaderBar(self)

        self.headerbar_stack.add_child(self.headerbar)
        self.headerbar_stack.props.visible_child = self.headerbar

        self._select_entry_status_page.props.icon_name = f"{const.APP_ID}-symbolic"

        self.active_element: SafeElement | None = None

        # Browser Mode
        self.browsing_panel = BrowsingPanel(self)
        self.toolbar_view.props.content = self.browsing_panel

        self.setup()

        self.clipboard = self.get_clipboard()

        self.db_locked_handler = self.database_manager.connect(
            "notify::locked",
            self._on_database_lock_changed,
        )

        # Sets the menu's save button sensitive property.
        save_action = window.lookup_action("db.save_dirty")
        dbm.bind_property("is-dirty", save_action, "enabled")

        search_action = Gio.PropertyAction.new("db.search", self, "search-active")
        window.add_action(search_action)

        def on_locked(dbm, _spec):
            if dbm.locked:
                self._search_bar.set_key_capture_widget(None)
            else:
                self._search_bar.set_key_capture_widget(window)

        dbm.connect("notify::locked", on_locked)
        self._search_bar.set_key_capture_widget(window)

        self._search_entry.connect("search-changed", self._on_search_changed)
        self._search_entry.connect("next-match", self._on_search_next_match)
        self._search_entry.connect("previous-match", self._on_search_previous_match)
        self.bind_property(
            "search-active",
            self._search_bar,
            "search-mode-enabled",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )

        self._split_view.connect("notify::show-content", self._on_show_content_notify)
        self._split_view.connect("notify::collapsed", self._on_collapsed_notify)

        self._selection_manager = SelectionManager(self)

    def inner_dispose(self):
        logging.debug("Database disposed")
        self.cleanup()

        if self.db_locked_handler:
            self.database_manager.disconnect(self.db_locked_handler)
            self.db_locked_handler = None

    def setup(self):
        self.start_save_loop()
        self.start_database_lock_timer()

        app = self.window.props.application
        self.session_handler_id = app.connect(
            "notify::screensaver-active",
            self.on_session_lock,
        )

    def listbox_row_factory(self, element: SafeElement) -> Gtk.Widget:
        if element.is_entry:
            row = EntryRow(self)
            row.props.safe_entry = element
            return row

        row = GroupRow(self)
        row.props.safe_group = element
        return row

    def show_edit_page(self, element: SafeElement, new: bool = False) -> None:
        self.start_database_lock_timer()
        # Sets the accessed time.
        element.touch()

        self._split_view.props.show_content = True

        if self.active_element == element:
            return

        self.active_element = element

        if element.is_group:
            page = GroupPage(self, element)
        else:
            page = EntryPage(self, element, new)

        page = Adw.NavigationPage.new(page, "Edit")
        self._split_view.props.content = page

        self.browsing_panel.selection_model.props.selected_item = element

    def show_browser_page(self, group: SafeGroup) -> None:
        self.start_database_lock_timer()

        self.active_element = group
        self.props.current_element = group
        self._split_view.props.show_content = False
        self.browsing_panel.visit_group(group)

    def _navigate_back(self):
        if element := self.active_element:
            parent_group = element.parentgroup
            self.show_browser_page(parent_group)
            return

        self.show_browser_page(self.current_element)

    #
    # Group and Entry Management
    #

    @GObject.Property(type=SafeElement)
    def current_element(self) -> SafeElement:
        # FIXME Rename to current_group, check all uses
        return self._current_element

    @current_element.setter  # type: ignore
    def current_element(self, element: SafeElement) -> None:
        self._current_element = element

    #
    # Events
    #

    def on_save(
        self,
        database_manager: DatabaseManager,
        result: Gio.AsyncResult,
    ) -> None:
        try:
            is_saved = database_manager.save_finish(result)
        except GLib.Error:
            logging.exception("Could not save Safe")
            self.window.send_notification(_("Could not save Safe"))
        else:
            if is_saved:
                self.window.send_notification(_("Safe saved"))
            else:
                self.window.send_notification(_("No changes made"))

    def on_auto_save(
        self,
        database_manager: DatabaseManager,
        result: Gio.AsyncResult,
    ) -> None:
        # TODO Does it make sense to present this info to the user?
        try:
            database_manager.save_finish(result)
        except GLib.Error:
            logging.exception("Could not automatically save Safe")

    def lock_safe(self):
        self.database_manager.props.locked = True

    def on_add_entry_action(self) -> None:
        """CB when the Add Entry menu was clicked."""
        group = self.props.current_element
        new_entry: SafeEntry = group.new_entry()
        self.show_edit_page(new_entry, new=True)

    def on_add_group_action(self) -> None:
        """CB when menu entry Add Group is clicked."""
        group = self.props.current_element
        safe_group = group.new_subgroup()
        self.show_edit_page(safe_group)

    def on_element_delete_action(self) -> None:
        """Delete the visible entry from the menu."""
        if element := self.active_element:
            parent_group = element.parentgroup
            self.browsing_panel.unselect(element)
            self.active_element = parent_group
            if element.trash():
                elements = []
            else:
                elements = [(element, parent_group)]

            self._split_view.props.content = self._empty_page
            self.deleted_notification(elements)
            self._navigate_back()

    def deleted_notification(self, elements):
        n_del = len(elements)
        if n_del:
            label = ngettext(
                "Element deleted",
                "{} Elements deleted",
                n_del,
            ).format(n_del)
        else:
            label = _("Deletion completed")

        toast = Adw.Toast.new(label)

        if n_del:
            toast.props.button_label = _("Undo")
            toast.props.action_name = "win.db.undo_delete"

        if data := self.undo_data:
            data.toast.dismiss()

        def dismissed_db(toast):
            if (data := self.undo_data) and data.toast == toast:
                self.undo_data = None

        toast.connect("dismissed", dismissed_db)

        self.undo_data = UndoData(elements, toast)
        self.window.toast_overlay.add_toast(toast)

    def undo_delete(self):
        if data := self.undo_data:
            for element, element_parent in data.elements:
                element.move_to(element_parent)

            self.undo_data = None

    def attribute_deleted(self, entry, key, value, protected):
        toast = Adw.Toast.new(_("Attribute deleted"))
        toast.props.button_label = _("Undo")
        toast.props.action_name = "win.db.undo_attribute_delete"

        if data := self.attribute_undo_data:
            data.toast.dismiss()

        undo_data = AttributeUndoData(toast, entry, key, value, protected)
        self.attribute_undo_data = undo_data
        self.window.toast_overlay.add_toast(toast)

    def undo_attribute_delete(self):
        if data := self.attribute_undo_data:
            data.toast.dismiss()
            data.entry.set_attribute(data.key, data.value, data.protected)

            self.attribute_undo_data = None

    def on_entry_duplicate_action(self):
        if element := self.active_element:
            element.duplicate()

    def send_to_clipboard(
        self,
        text: str,
        message: str = "",
        toast_overlay: Adw.ToastOverlay | None = None,
    ) -> None:
        if not message:
            message = _("Copied")

        self.start_database_lock_timer()

        if self.on_is_active_handler:
            self.disconnect(self.on_is_active_handler)
            self.on_is_active_handler = None

        self.clipboard.set(text)

        if toast_overlay:
            window = toast_overlay.get_root()
            toast = Adw.Toast.new(message)
            toast_overlay.add_toast(toast)
        else:
            window = self.window
            self.window.send_notification(message)

        clear_clipboard_time = gsecrets.config_manager.get_clear_clipboard()
        clipboard_last_content = self.clipboard.get_content()

        def callback():
            if window.props.is_active:
                if clipboard_last_content == self.clipboard.get_content():
                    logging.debug("Clipboard cleared")
                    self.clipboard.set_content(None)
            else:
                logging.debug(
                    "Could not clear the clipboard, it will clear on new focus",
                )

                def on_is_active(window, _pspec):
                    if self.on_is_active_handler:
                        window.disconnect(self.on_is_active_handler)
                        self.on_is_active_handler = None

                    # Only clear clipboard if we set ourselves the content.
                    if clipboard_last_content == self.clipboard.get_content():
                        logging.debug("Clipboard cleared")
                        self.clipboard.set_content(None)

                self.on_is_active_handler = window.connect(
                    "notify::is-active",
                    on_is_active,
                )

            return GLib.SOURCE_REMOVE

        GLib.timeout_add_seconds(clear_clipboard_time, callback)

    def show_database_settings(self) -> None:
        DatabaseSettingsDialog(self).present(self)

    def on_session_lock(self, app: Gtk.Application, _pspec: GObject.ParamSpec) -> None:
        if (
            app.props.screensaver_active
            and not self.database_manager.props.locked
            and gsecrets.config_manager.get_lock_on_session_lock()
        ):
            self.lock_timeout_database()

    #
    # Dialog Creator
    #

    def show_references_dialog(self) -> None:
        """Show a Group/Entry reference dialog.

        Invoked by the app.entry.references action
        """
        ReferencesDialog(self).present(self)

    def show_properties_dialog(self) -> None:
        """Show a Group/Entry property dialog.

        Invoked by the app.element.properties action
        """
        PropertiesDialog(self).present(self)

    #
    # Utils
    #

    def _on_database_lock_changed(self, _database_manager, _value):
        locked = self.database_manager.props.locked
        if locked:
            self.cleanup()
            if gsecrets.config_manager.get_save_automatically():
                self.auto_save_database()

            filepath = self.database_manager.path
            self.window.start_database_opening_routine(filepath)
        else:
            self.window.view = self.window.View.UNLOCKED_DATABASE
            self.setup()

    def lock_timeout_database(self):
        self.database_manager.props.locked = True

        # NOTE: Notification that a safe has been locked.
        self.window.send_notification(_("Safe locked due to inactivity"))

    #
    # Helper Methods
    #

    def cleanup(self) -> None:
        """Stop all ongoing operations.

        * stop the save loop
        * cancel all timers
        * stop lisening to screensaver

        This is the opposite of setup().
        """
        logging.debug("Cleaning database %s", self.database_manager.path)

        if self._lock_timer_handler:
            GLib.source_remove(self._lock_timer_handler)
            self._lock_timer_handler = None

        # Do not listen to screensaver kicking in anymore
        if self.session_handler_id:
            app = self.window.props.application
            if app:
                app.disconnect(self.session_handler_id)

            self.session_handler_id = None

        # stop the save loop
        if self.save_loop:
            GLib.source_remove(self.save_loop)
            self.save_loop = None

        # Cleanup temporal files created when opening attachments.
        def callback(gfile, result):
            try:
                gfile.delete_finish(result)
            except GLib.Error:
                logging.exception("Could not delete temporal file")

        cache_dir = Path(GLib.get_user_cache_dir()) / const.SHORT_NAME / "tmp"
        for root, _dirs, files in os.walk(cache_dir):
            for file_name in files:
                file_path = Path(root) / file_name
                gfile = Gio.File.new_for_path(str(file_path))
                gfile.delete_async(GLib.PRIORITY_DEFAULT, None, callback)

    def save_database(self) -> None:
        """Save the database.

        Shows a notification after saving.
        """

        def on_check_file_changes(dbm, result):
            self.on_check_file_changes(dbm, result, self.on_save)

        self.database_manager.check_file_changes_async(on_check_file_changes)

    def auto_save_database(self) -> None:
        """Save the database."""
        logging.debug("Automatically saving database")

        def on_check_file_changes(dbm, result):
            self.on_check_file_changes(dbm, result, self.on_auto_save)

        self.database_manager.check_file_changes_async(on_check_file_changes)

    def on_check_file_changes(self, dbm, result, on_save_callback):
        try:
            conflicts = dbm.check_file_changes_finish(result)
        except GLib.Error:
            logging.exception("Could not monitor file changes")
            self.window.send_notification(_("Could not save Safe"))
        else:
            if conflicts:
                dialog = SavingConflictDialog(self.window, dbm, on_save_callback)
                dialog.present(self.window)
            else:
                self.database_manager.save_async(on_save_callback)

    def start_database_lock_timer(self):
        if self._lock_timer_handler:
            GLib.source_remove(self._lock_timer_handler)
            self._lock_timer_handler = None

        if self.database_manager.props.locked:
            return

        timeout = gsecrets.config_manager.get_database_lock_timeout() * 60
        if timeout:
            self._lock_timer_handler = GLib.timeout_add_seconds(
                timeout,
                self.lock_timeout_database,
            )

    def start_save_loop(self):
        logging.debug("Starting automatic save loop")
        self.save_loop = GLib.timeout_add_seconds(30, self.threaded_save_loop)

    def threaded_save_loop(self) -> bool:
        """Return True if the safe was saved."""
        if gsecrets.config_manager.get_save_automatically():
            self.auto_save_database()

        return GLib.SOURCE_CONTINUE

    def go_back(self):
        if self._search_active:
            self.props.search_active = False
            return

        if self._selection_mode:
            self.props.selection_mode = False
            return

        self._navigate_back()

    @Gtk.Template.Callback()
    def _on_selection_go_back_button_clicked(self, _button):
        self.headerbar.on_go_back_button_clicked(None)

    @GObject.Property(type=bool, default=False)
    def search_active(self) -> bool:
        """Property to know if search is active.

        It is used by Search to update the widgets accordingly.

        :returns: True is search is active
        :rtype: bool
        """
        return self._search_active

    @search_active.setter  # type: ignore
    def search_active(self, value: bool) -> None:
        """Set the search mode.

        :param value: new search_active
        """
        self._search_active = value

        if not value:
            self.browsing_panel.set_search(None)

    # TODO This property does not emit a notify signal when the manager locked
    # property is updated.
    @GObject.Property(type=bool, default=False)
    def database_locked(self):
        """Get database lock status.

        :returns: True if the database is locked
        :rtype: bool
        """
        return self.database_manager.props.locked

    @GObject.Property(type=bool, default=False)
    def selection_mode(self) -> bool:
        return self._selection_mode

    @selection_mode.setter  # type: ignore
    def selection_mode(self, value: bool) -> None:
        self._selection_mode = value
        self._selection_mode_action_bar.set_revealed(value)

        if value:
            self.headerbar_stack.props.visible_child = self._selection_mode_headerbar
        else:
            self.headerbar_stack.props.visible_child = self.headerbar
            self._clear_selection()

    @Gtk.Template.Callback()
    def _on_selection_cancel_clicked(self, _button):
        self.props.selection_mode = False

    @Gtk.Template.Callback()
    def _on_clear_selection_clicked(self, _button):
        self.start_database_lock_timer()
        self._clear_selection()

    @Gtk.Template.Callback()
    def _on_cut_selection_clicked(self, _button):
        self.start_database_lock_timer()
        self._selection_manager.cut_selection()
        self._update_selection()

    @Gtk.Template.Callback()
    def _on_paste_selection_clicked(self, _button):
        self.start_database_lock_timer()
        self._selection_manager.paste_selection()
        self._update_selection()
        self.props.selection_mode = False

    @Gtk.Template.Callback()
    def _on_delete_selection_clicked(self, _button):
        self.start_database_lock_timer()
        if (
            element := self.browsing_panel.selection_model.selected_item
        ) and element.selected:
            self._split_view.props.content = self._empty_page
            self._navigate_back()

        self._selection_manager.delete_selection()
        self._update_selection()
        self.props.selection_mode = False

    def add_selection(self, element):
        if element.is_group:
            self._selection_manager.add_group(element)
        else:
            self._selection_manager.add_entry(element)

        self._update_selection()

    def remove_selection(self, element):
        if element.is_group:
            self._selection_manager.remove_group(element)
        else:
            self._selection_manager.remove_entry(element)

        self._update_selection()

    def _clear_selection(self):
        self._selection_manager.clear_selection()
        self._update_selection()

    def _update_selection(self):
        n_selected = self._selection_manager.selected_elements
        cut_mode = self._selection_manager.cut_mode
        non_empty_selection = n_selected > 0

        self._cut_selection_button.props.sensitive = non_empty_selection
        self._delete_selection_button.props.sensitive = non_empty_selection
        self._clear_selection_button.props.sensitive = non_empty_selection

        if n_selected == 0:
            title = _("Select Items")
        else:
            # TRANSLATORS e.g. "21 Selected". What is selected is an 'Item'.
            title = ngettext("{} Selected", "{} Selected", n_selected).format(
                n_selected,
            )

        self._selection_mode_title.props.title = title

        self._delete_selection_button.props.sensitive = n_selected > 0

        if cut_mode:
            visible_child = self._cut_selection_button
        else:
            visible_child = self._paste_selection_button

        self._cut_paste_button_stack.props.visible_child = visible_child

    def _update_selection_on_collapsed(self):
        nav = self._split_view
        hide_selection = nav.props.collapsed and not nav.props.show_content
        self.browsing_panel.hide_selection(hide_selection)

    def _on_show_content_notify(self, _split_view, _pspec):
        self._update_selection_on_collapsed()

    def _on_collapsed_notify(self, _split_view, _pspec):
        self._update_selection_on_collapsed()

    def _on_search_changed(self, search_entry):
        self.browsing_panel.set_search(search_entry.props.text)

    def _on_search_next_match(self, _search_entry):
        self.browsing_panel.select_next()

    def _on_search_previous_match(self, _search_entry):
        self.browsing_panel.select_previous()



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import typing

from gi.repository import Adw, GObject, Gtk

from gsecrets.safe_element import EntryColor

if typing.TYPE_CHECKING:
    from gsecrets.database_manager import DatabaseManager
    from gsecrets.safe_element import SafeEntry
    from gsecrets.unlocked_database import UnlockedDatabase


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/color_button.ui")
class ColorButton(Gtk.FlowBoxChild):
    __gtype_name__ = "ColorButton"

    _button = Gtk.Template.Child()

    def __init__(self, color: EntryColor, is_selected: bool):
        """RadioButton to select the color of an entry.

        :param EntryColor color: color of the button
        :param bool selected: True if the color is selected
        """
        super().__init__()

        self._is_selected = False
        self._color: EntryColor = color

        self.props.is_selected = is_selected

        self._button.add_css_class(self._color.css_class())
        self._button.props.tooltip_text = color.to_translatable()

    @property
    def color(self) -> str:
        """Color of the widget."""
        return self._color.value

    @GObject.Property(type=bool, default=False)
    def is_selected(self) -> bool:
        return self._is_selected

    @is_selected.setter  # type: ignore
    def is_selected(self, is_selected: bool) -> None:
        self._is_selected = is_selected
        if is_selected:
            self._button.props.icon_name = "checkmark-symbolic"
        else:
            self._button.props.icon_name = ""

    def connect_clicked(self, callback):
        def cb_wrapper(_button, color_button):
            callback(color_button)

        self._button.connect("clicked", cb_wrapper, self)


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/color_entry_row.ui")
class ColorEntryRow(Adw.PreferencesGroup):
    __gtype_name__ = "ColorEntryRow"

    _flowbox = Gtk.Template.Child()

    def __init__(self, unlocked_database: UnlockedDatabase, safe_entry: SafeEntry):
        """Widget to select the color of an entry.

        :param UnlockedDatabase unlocked_database: unlocked database
        :param SafeEntry safe_entry: the safe entry
        """
        super().__init__()

        self._unlocked_database: UnlockedDatabase = unlocked_database
        self._db_manager: DatabaseManager = unlocked_database.database_manager
        self._safe_entry: SafeEntry = safe_entry

        for color in EntryColor:
            is_selected: bool = safe_entry.props.color == color.value
            color_button: ColorButton = ColorButton(color, is_selected)
            self._flowbox.append(color_button)
            color_button.connect_clicked(self._on_color_clicked)

    def _on_color_clicked(
        self,
        selected_child: ColorButton,
    ) -> None:
        if selected_child.props.is_selected:
            return

        for child in self._flowbox:  # pylint: disable=not-an-iterable
            selected: bool = child == selected_child
            child.props.is_selected = selected

        self._safe_entry.props.color = selected_child.color



# SPDX-License-Identifier: GPL-3.0-only
"""Responsible for displaying the Entry/Group Properties."""

from __future__ import annotations

from gi.repository import Adw, Gtk

from gsecrets.utils import format_time


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/properties_dialog.ui")
class PropertiesDialog(Adw.Dialog):
    """Displays a modal dialog with Entry/Group Properties."""

    __gtype_name__ = "PropertiesDialog"

    _accessed_row = Gtk.Template.Child()
    _created_row = Gtk.Template.Child()
    _modified_row = Gtk.Template.Child()
    _uuid_row = Gtk.Template.Child()

    def __init__(self, database):
        super().__init__()

        self.__database = database
        self.__db_manager = database.database_manager
        self.__signal_handle = 0

        self.__setup_signals()
        self.__setup_widgets()

    def __update_properties(self) -> None:
        """Construct dialog content with the attributes of the Entry|Group."""
        element = self.__database.current_element
        hex_uuid = element.uuid.hex.upper()
        self._uuid_row.props.subtitle = hex_uuid
        self._accessed_row.props.subtitle = format_time(element.atime)
        self._modified_row.props.subtitle = format_time(element.mtime)
        self._created_row.props.subtitle = format_time(element.ctime)

    def __setup_signals(self) -> None:
        handle = self.__db_manager.connect("notify::locked", self.__on_locked)
        self.__signal_handle = handle

    def __setup_widgets(self) -> None:
        self.__update_properties()

    def __on_locked(self, database_manager, _value):
        locked = database_manager.props.locked
        if locked:
            self.close()

    def do_closed(self) -> None:
        if handle := self.__signal_handle:
            self.__db_manager.disconnect(handle)
            self.__signal_handle = 0



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from gettext import gettext as _

from gi.repository import Adw, Gtk


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/add_attribute_dialog.ui")
class AddAttributeDialog(Adw.Dialog):
    # pylint: disable=too-many-instance-attributes

    __gtype_name__ = "AddAttributeDialog"

    _key_entry = Gtk.Template.Child()
    _protected_switch = Gtk.Template.Child()
    _value_entry = Gtk.Template.Child()
    _toast_overlay = Gtk.Template.Child()

    def __init__(self, db_manager, entry):
        super().__init__()

        self.entry = entry
        self._db_manager = db_manager

        self._signal_handle = db_manager.connect("notify::locked", self._on_locked)

    @Gtk.Template.Callback()
    def _on_add_button_clicked(self, _button):
        key = self._key_entry.props.text

        # TODO Remove once https://github.com/libkeepass/pykeepass/issues/254 is
        # fixed
        if '"' in key or "'" in key:
            self._key_entry.add_css_class("error")
            toast = Adw.Toast.new(_("Attribute key contains an illegal character"))
            self._toast_overlay.add_toast(toast)
            return

        if key == "" or key is None:
            self._key_entry.add_css_class("error")
            return

        if self.entry.has_attribute(key):
            self._key_entry.add_css_class("error")
            toast = Adw.Toast.new(_("Attribute key already exists"))
            self._toast_overlay.add_toast(toast)
            return

        value = self._value_entry.props.text
        protected = self._protected_switch.props.active

        self.entry.set_attribute(key, value, protected)
        self.close()

    def _on_locked(self, database_manager, _value):
        locked = database_manager.props.locked
        if locked:
            self.close()

    def do_closed(self) -> None:
        if handle := self._signal_handle:
            self._db_manager.disconnect(handle)
            self._signal_handle = 0



# SPDX-License-Identifier: GPL-3.0-only
from gi.repository import Adw, GObject, Gtk


class OsdProgressBar(Adw.Bin):
    __gtype_name__ = "OsdProgressBar"

    def __init__(self):
        super().__init__()

        progress = Gtk.ProgressBar(
            visible=False,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.START,
            pulse_step=0.01,
        )
        progress.add_css_class("osd")

        overlay = Gtk.Overlay()
        overlay.add_overlay(progress)

        self.props.child = overlay

        self._progress = progress
        self._overlay = overlay

        def callback(_value: float) -> None:
            self._progress.pulse()

        target = Adw.CallbackAnimationTarget.new(callback)
        animation = Adw.TimedAnimation.new(
            self,
            0.0,
            0.0,
            Adw.DURATION_INFINITE,
            target,
        )

        self._animation = animation

    def start_pulse(self) -> None:
        self._progress.set_visible(True)
        self._animation.play()

    def end_pulse(self) -> None:
        self._progress.set_visible(False)
        self._animation.pause()
        self._animation.reset()

    @GObject.Property(type=Gtk.Widget)
    def content(self):
        return self._overlay.props.child

    @content.setter  # type: ignore
    def content(self, content):
        self._overlay.props.child = content



# SPDX-License-Identifier: GPL-3.0-only
# pylint: skip-file

from gi.repository import GObject

from gsecrets.password_generator_popover import PasswordGeneratorPopover
from gsecrets.unlocked_headerbar import UnlockedHeaderBar
from gsecrets.welcome_page import WelcomePage
from gsecrets.widgets.add_list_box import AddListBox
from gsecrets.widgets.credentials_group import CredentialsGroup
from gsecrets.widgets.error_revealer import ErrorRevealer
from gsecrets.widgets.expiration_date_row import ExpirationDateRow
from gsecrets.widgets.history_row import HistoryRow
from gsecrets.widgets.history_window import HistoryWindow
from gsecrets.widgets.locked_headerbar import LockedHeaderBar
from gsecrets.widgets.notes_dialog import NotesDialog
from gsecrets.widgets.osd_progress_bar import OsdProgressBar
from gsecrets.widgets.password_level_bar import PasswordLevelBar
from gsecrets.widgets.progress_icon import ProgressIcon


def load_widgets():
    GObject.type_ensure(CredentialsGroup)
    GObject.type_ensure(PasswordGeneratorPopover)
    GObject.type_ensure(UnlockedHeaderBar)
    GObject.type_ensure(WelcomePage)
    GObject.type_ensure(AddListBox)
    GObject.type_ensure(ErrorRevealer)
    GObject.type_ensure(ExpirationDateRow)
    GObject.type_ensure(HistoryRow)
    GObject.type_ensure(HistoryWindow)
    GObject.type_ensure(LockedHeaderBar)
    GObject.type_ensure(NotesDialog)
    GObject.type_ensure(OsdProgressBar)
    GObject.type_ensure(PasswordLevelBar)
    GObject.type_ensure(ProgressIcon)
